import httplib
import json
import re

import requests

from cumulusci.core.exceptions import GithubApiNotFoundError
from cumulusci.core.exceptions import GithubApiNoResultsError
from cumulusci.core.exceptions import GithubApiUnauthorized


class GithubApiMixin(object):
    github_api_base_url = 'https://api.github.com'

    @property
    def github_owner(self):
        return self.github_info['github_owner']

    @property
    def github_repo(self):
        return self.github_info['github_repo']

    @property
    def github_username(self):
        return self.github_info['github_username']

    @property
    def github_password(self):
        return self.github_info['github_password']

    @property
    def master_branch(self):
        return self.github_info.get('master_branch', 'master')

    @property
    def prefix_prod(self):
        return self.github_info.get('prefix_prod', 'prod/')

    @property
    def prefix_beta(self):
        return self.github_info.get('prefix_beta', 'beta/')

    @property
    def current_tag_info(self):
        is_prod = False
        is_beta = False
        tag = self.current_tag
        if tag.startswith(self.prefix_prod):
            is_prod = True
        elif tag.startswith(self.prefix_beta):
            is_beta = True

        if is_prod:
            version_number = tag.replace(self.prefix_beta, '')
        elif is_beta:
            version_parts = re.findall(
                '{}(\d+\.\d+)-Beta_(\d+)'.format(self.prefix_beta),
                tag,
            )
            assert version_parts
            version_number = '{} (Beta {})'.format(*version_parts[0])
        else:
            version_number = None

        tag_info = {
            'is_prod': is_prod,
            'is_beta': is_beta,
            'version_number': version_number,
        }

        return tag_info

    def call_api(self, subpath, data=None):
        """ Takes a subpath under the repository (ex: /releases) and returns
        the json data from the api """
        api_url = '{}/repos/{}/{}{}'.format(
            self.github_api_base_url, self.github_owner, self.github_repo,
            subpath)

        # Use Github Authentication if available for the repo
        kwargs = {}
        if self.github_owner and self.github_owner:
            kwargs['auth'] = (self.github_owner, self.github_password)

        if data:
            resp = requests.post(api_url, data=json.dumps(data), **kwargs)
        else:
            resp = requests.get(api_url, **kwargs)

        if resp.status_code == httplib.NOT_FOUND:
            raise GithubApiNotFoundError(resp.content)
        elif resp.status_code == httplib.UNAUTHORIZED:
            raise GithubApiUnauthorized(resp.content)

        try:
            data = json.loads(resp.content)
            return data
        except:
            return resp.status_code



from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
import http.client
import threading
import time
import unittest
import urllib.request, urllib.error, urllib.parse

import mock
import responses

from cumulusci.oauth.salesforce import CaptureSalesforceOAuth
from cumulusci.oauth.exceptions import SalesforceOAuthError


@mock.patch('webbrowser.open', mock.MagicMock(return_value=None))
class TestCaptureSalesforceOAuth(unittest.TestCase):

    def _create_oauth(self):
        return CaptureSalesforceOAuth(
            self.client_id,
            self.client_secret,
            self.callback_url,
            self.auth_site,
            self.scope,
        )

    def setUp(self):
        self.client_id = 'foo_id'
        self.client_secret = 'foo_secret'
        self.callback_url = 'http://localhost:8080'
        self.scope = 'refresh_token web full'
        self.auth_site = 'https://login.salesforce.com'

    @responses.activate
    def test_oauth_flow(self):

        # mock response to URL validation
        responses.add(
            responses.GET,
            'https://login.salesforce.com/services/oauth2/authorize',
            status=http.client.OK,
        )

        # mock response for SalesforceOAuth2.get_token()
        expected_response = {
            u'access_token': u'abc123',
            u'id_token': u'abc123',
            u'token_type': u'Bearer',
            u'signature': u'abc123',
            u'issued_at': u'12345',
            u'scope': u'{}'.format(self.scope),
            u'instance_url': u'https://na15.salesforce.com',
            u'id': u'https://login.salesforce.com/id/abc/xyz',
            u'refresh_token': u'abc123',
        }
        responses.add(
            responses.POST,
            'https://login.salesforce.com/services/oauth2/token',
            status=http.client.OK,
            json=expected_response,
        )

        # create CaptureSalesforceOAuth instance
        o = self._create_oauth()

        # call OAuth object on another thread - this spawns local httpd
        t = threading.Thread(target=o.__call__)
        t.start()
        while True:
            if o.httpd:
                break
            print('waiting for o.httpd')
            time.sleep(0.01)

        # simulate callback from browser
        response = urllib.request.urlopen(self.callback_url + '?code=123')

        # wait for thread to complete
        t.join()

        # verify
        self.assertEqual(o.response.json(), expected_response)
        self.assertEqual(response.read(), b'OK')

    @responses.activate
    def test_bad_request(self):

        # mock response to URL validation
        responses.add(
            responses.GET,
            'https://login.salesforce.com/services/oauth2/authorize',
            status=http.client.BAD_REQUEST,
        )

        # create CaptureSalesforceOAuth instance
        o = self._create_oauth()

        # call OAuth object with bad info
        with self.assertRaises(SalesforceOAuthError):
            o()


if __name__ == '__main__':
    unittest.main()



from cumulusci.tasks.salesforce import BaseSalesforceApiTask

import csv
import time
import hiyapyco
import xml.etree.ElementTree as ET

import datetime
import requests
import tempfile
import unicodecsv

from collections import OrderedDict

from salesforce_bulk import CsvDictsAdapter

from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import create_session
from sqlalchemy.orm import mapper
from sqlalchemy.orm import Session
from sqlalchemy import create_engine
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import Unicode
from sqlalchemy import text
from sqlalchemy import types
from sqlalchemy import event
from StringIO import StringIO

# TODO: UserID Catcher
# TODO: Dater

# Create a custom sqlalchemy field type for sqlite datetime fields which are stored as integer of epoch time
class EpochType(types.TypeDecorator):
    impl = types.Integer

    epoch = datetime.datetime(1970, 1, 1, 0, 0, 0)

    def process_bind_param(self, value, dialect):
        return (value / 1000 - self.epoch).total_seconds()

    def process_result_value(self, value, dialect):
        return self.epoch + datetime.timedelta(seconds=value / 1000)

# Listen for sqlalchemy column_reflect event and map datetime fields to EpochType
@event.listens_for(Table, "column_reflect")
def setup_epoch(inspector, table, column_info):
    if isinstance(column_info['type'], types.DateTime):
        column_info['type'] = EpochType()

class DeleteData(BaseSalesforceApiTask):

    task_options = {
        'objects': {
            'description': 'A list of objects to delete records from in order of deletion.  If passed via command line, use a comma separated string',
            'required': True,
        }
    }

    def _init_options(self, kwargs):
        super(DeleteData, self)._init_options(kwargs)
       
        # Split and trim objects string into a list if not already a list
        if not isinstance(self.options['objects'], list):
            self.options['objects'] = [obj.strip() for obj in self.options['objects'].split(',')]

    def _run_task(self):
        for obj in self.options['objects']:
            self.logger.info('Deleting all {} records'.format(obj))
            # Query for all record ids
            self.logger.info('  Querying for all {} objects'.format(obj))
            query_job = self.bulk.create_query_job(obj, contentType='CSV')
            batch = self.bulk.query(query_job, "select Id from {}".format(obj))
            while not self.bulk.is_batch_done(batch, query_job):
                time.sleep(10)
            self.bulk.close_job(query_job)
            delete_rows = []
            for result in self.bulk.get_all_results_for_query_batch(batch,query_job):
                reader = unicodecsv.DictReader(result, encoding='utf-8')
                for row in reader:
                    delete_rows.append(row)

            if not delete_rows:
                self.logger.info('  No {} objects found, skipping delete'.format(obj))
                continue

            # Delete the records
            delete_job = self.bulk.create_delete_job(obj, contentType='CSV')
            self.logger.info('  Deleting {} {} records'.format(len(delete_rows), obj))
            batch_num = 1
            for batch in self._upload_batch(delete_job, delete_rows):
                self.logger.info('    Uploaded batch {}'.format(batch))
                while not self.bulk.is_batch_done(batch, delete_job):
                    self.logger.info('      Checking status of batch {0}'.format(batch_num))
                    time.sleep(10)
                self.logger.info('      Batch {} complete'.format(batch))
                batch_num += 1
            self.bulk.close_job(delete_job)

    def _split_batches(self, data, batch_size):
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(data), batch_size):
            yield data[i:i + batch_size]

    def _upload_batch(self, job, data):
        # Split into batches
        batches = self._split_batches(data, 2500)

        uri = "{}/job/{}/batch".format(self.bulk.endpoint, job)
        headers = self.bulk.headers({"Content-Type": "text/csv"})
        for batch in batches:
            data = ['"Id"']
            data += ['"{}"'.format(record['Id']) for record in batch]
            data = '\n'.join(data)
            resp = requests.post(uri, data=data, headers=headers)
            content = resp.content

            if resp.status_code >= 400:
                self.bulk.raise_error(content, resp.status_code)

            tree = ET.fromstring(content)
            batch_id = tree.findtext("{%s}id" % self.bulk.jobNS)

            yield batch_id
        
class LoadData(BaseSalesforceApiTask):

    task_options = {
        'database_url': {
            'description': 'The database url to a database containing the test data to load',
            'required': True,
        },
        'mapping': {
            'description': 'The path to a yaml file containing mappings of the database fields to Salesforce object fields',
            'required': True,
        },
    }

    def _run_task(self):
        self._init_mapping()
        self._init_db()

        for name, mapping in self.mapping.items():
            api = mapping.get('api', 'bulk')
            if mapping.get('retrieve_only', False):
                continue

            self.logger.info('Running Job: {} with {} API'.format(name, api))
            rows = self._get_batches(mapping)

            if api is 'bulk':
                self._upload_batches(mapping, rows)
            elif api is 'sobject':
                self._sobject_api_upload_batches(mapping, rows)

    def _sobject_api_upload_batches(self, mapping, batches):
        for batch, batch_rows in batches:
            pass

    def _create_job(self, mapping):
        action = mapping.get('action', 'insert')
        job_id = None

        if action == 'insert':
            job_id = self.bulk.create_insert_job(mapping['sf_object'], contentType='CSV')

        if not job_id:
            self.logger.error('  No handler for action type {}'.format(action))

        self.logger.info('  Created bulk job {}'.format(job_id))

        return job_id

    def _upload_batches(self, mapping, batches):
        job_id = None
        table = self.tables[mapping.get('table')]

        for batch, batch_rows in batches:
            if not job_id:
                # Create a job only once we have the first batch to load into it
                job_id = self._create_job(mapping)

            # Prepare the rows
            rows = CsvDictsAdapter(iter(batch))

            # Create the batch
            batch_id = self.bulk.post_batch(job_id, rows)
            self.logger.info('    Uploaded batch {}'.format(batch_id))
            while not self.bulk.is_batch_done(batch_id, job_id):
                self.logger.info('      Checking batch status...')
                time.sleep(10)
            
            # Wait for batch to complete
            res = self.bulk.wait_for_batch(job_id, batch_id)
            self.logger.info('      Batch {} complete'.format(batch_id))

            # salesforce_bulk is broken in fetching id results so do it manually
            results_url = '{}/job/{}/batch/{}/result'.format(self.bulk.endpoint, job_id, batch_id)
            headers = self.bulk.headers()
            resp = requests.get(results_url, headers=headers)
            csv_file = tempfile.TemporaryFile()
            csv_file.write(resp.content)
            csv_file.seek(0)
            reader = csv.DictReader(csv_file)

            # Write to the local Id column on the uploaded rows
            i = 0
            for result in reader:
                row = batch_rows[i]
                i += 1
                if result['Id']:
                    setattr(row, mapping['fields']['Id'], result['Id'])

            # Commit to the db
            self.session.commit()
                
        
        self.bulk.close_job(job_id)
        status = self.bulk.job_status(job_id)
        
    def _query_db(self, mapping):
        table = self.tables[mapping.get('table')]

        query = self.session.query(table)
        if 'filters' in mapping:
            filter_args = []
            for f in mapping['filters']:
                filter_args.append(text(f))
            query = query.filter(*filter_args)
        return query


    def _get_batches(self, mapping, batch_size=None):
        if batch_size is None:
            batch_size = 10000

        action = mapping.get('action', 'insert')
        fields = mapping.get('fields', {}).copy()
        static = mapping.get('static', {})
        lookups = mapping.get('lookups', {})
        record_type = mapping.get('record_type')

        # Skip Id field on insert
        if action == 'insert' and 'Id' in fields:
            del fields['Id']

        # Build the list of fields to import
        import_fields = fields.keys() + static.keys() + lookups.keys()

        if record_type:
            import_fields.append('RecordTypeId')
            # default to the profile assigned recordtype if we can't find any
            # query for the RT by developer name
            try:
                query = "SELECT Id FROM RecordType WHERE SObjectType='{0}'" \
                    "AND DeveloperName = '{1}' LIMIT 1"
                record_type_id = self.sf.query(
                    query.format(mapping.get('sf_object'), record_type)
                )['records'][0]['Id']
            except (KeyError, IndexError):
                record_type_id = None

        query = self._query_db(mapping)

        total_rows = 0
        batch_num = 1
        batch = []
        batch_rows = []

        for row in query:
            total_rows += 1

            # Get the row data from the mapping and database values
            csv_row = {}
            for key, value in fields.items():
                csv_row[key] = getattr(row, value)
            for key, value in static.items():
                csv_row[key] = value
            for key, lookup in lookups.items():
                kwargs = {lookup['join_field']: getattr(row, lookup['key_field'])}
                try:
                    res = self.session.query(self.tables[lookup['table']]).filter_by(**kwargs).one()
                    csv_row[key] = getattr(res, lookup['value_field'])
                except:
                    csv_row[key] = None
            if record_type:
                csv_row['RecordTypeId'] = record_type_id

            # utf-8 encode row values
            for key, value in csv_row.items():
                if value:
                    if isinstance(value, datetime.datetime):
                        csv_row[key] = value.isoformat()
                    else:
                        try:
                            csv_row[key] = value.encode('utf8')
                        except AttributeError:
                            continue

            # Write to csv file
            batch.append(csv_row)

            batch_rows.append(row)

            # Slice into batches
            if len(batch) == batch_size:
                self.logger.info('    Processing batch {}'.format(batch_num))
                yield batch, batch_rows

                # Start the next batch
                batch_num += 1
                batch = []
                batch_rows = []

        self.logger.info('  Prepared {} rows for import to {}'.format(total_rows, mapping['sf_object']))

        if batch:
            yield batch, batch_rows

    def _init_db(self):
        # initialize the DB engine
        self.engine = create_engine(self.options['database_url'])

        # initialize DB metadata
        self.metadata = MetaData()
        self.metadata.bind = self.engine

        # initialize the automap mapping
        self.base = automap_base(bind=self.engine, metadata=self.metadata)
        self.base.prepare(self.engine, reflect=True)

        # Loop through mappings and reflect each referenced table
        self.tables = {}
        for name, mapping in self.mapping.items():
            if 'table' in mapping and mapping['table'] not in self.tables:
                self.tables[mapping['table']] = self.base.classes[mapping['table']]

        # initialize the DB session
        self.session = Session(self.engine)

    def _init_mapping(self):
        self.mapping = hiyapyco.load(self.options['mapping'])

class QueryData(BaseSalesforceApiTask):
    task_options = {
        'database_url': {
            'description': 'A DATABASE_URL where the query output should be written',
            'required': True,
        },
        'mapping': {
            'description': 'The path to a yaml file containing mappings of the database fields to Salesforce object fields',
            'required': True,
        },
    }

    def _run_task(self):
        self._init_mapping()
        self._init_db()

        for name, mapping in self.mappings.items():
            fields = self._fields_for_mapping(mapping)
            soql = self._soql_for_mapping(mapping)
            self._run_query(soql, mapping)

    def _init_db(self):
        self.models = {}

        # initialize the DB engine
        self.engine = create_engine(self.options['database_url'])

        # initialize DB metadata
        self.metadata = MetaData()
        self.metadata.bind = self.engine

        # Create the tables
        self._create_tables()

        # initialize the automap mapping
        self.base = automap_base(bind=self.engine, metadata=self.metadata)
        self.base.prepare(self.engine, reflect=True)

        # Loop through mappings and reflect each referenced table
        self.tables = {}
        #for name, mapping in self.mapping.items():
            #if 'table' in mapping and mapping['table'] not in self.tables:
                #self.tables[mapping['table']] = self.base.classes[mapping['table']]

        # initialize session
        self.session = create_session(bind=self.engine, autocommit=False)

    def _init_mapping(self):
        self.mappings = hiyapyco.load(self.options['mapping'])
        #self.mappings = [(name, mapping) for name, mapping in self.mappings.items()]
        #self.mappings.reverse()
        #rev_mappings = OrderedDict()
        #for mapping_item in self.mappings:
        #    rev_mappings[mapping_item[0]] = mapping_item[1]
        #self.mappings = rev_mappings

    def _soql_for_mapping(self, mapping):
        sf_object = mapping['sf_object']
        fields = [field['sf'] for field in self._fields_for_mapping(mapping)]
        soql = "SELECT {fields} FROM {sf_object}".format(**{
            'fields': ', '.join(fields),
            'sf_object': sf_object,
        })
        if 'record_type' in mapping:
            soql += ' WHERE RecordType.DeveloperName = \'{}\''.format(mapping['record_type'])
        return soql

    def _run_query(self, soql, mapping):
        self.logger.info('Creating bulk job for: {sf_object}'.format(**mapping))
        job = self.bulk.create_query_job(mapping['sf_object'], contentType='CSV')
        self.logger.info('Job id: {0}'.format(job))
        self.logger.info('Submitting query: {}'.format(soql))
        batch = self.bulk.query(job, soql)
        self.logger.info('Batch id: {0}'.format(batch))
        self.bulk.wait_for_batch(job, batch)
        self.logger.info('Batch {0} finished'.format(batch))
        self.bulk.close_job(job)
        self.logger.info('Job {0} closed'.format(job))

        field_map = {}
        for field in self._fields_for_mapping(mapping):
            field_map[field['sf']] = field['db']

        for result in self.bulk.get_all_results_for_query_batch(batch, job):
            reader = unicodecsv.DictReader(result, encoding='utf-8')
            for row in reader:
                self._import_row(row, mapping, field_map)

        self.session.commit()

    def _import_row(self, row, mapping, field_map):
        model = self.models[mapping['table']]
        mapped_row = {}
        for key, value in row.items():
            if key in mapping.get('lookups', {}):
                if not value:
                    mapped_row[field_map[key]] = None
                    continue
                # For lookup fields, the value should be the local db id instead of the sf id
                lookup = mapping['lookups'][key]
                lookup_model = self.models[lookup['table']]
                kwargs = { lookup['value_field']: value }
                res = self.session.query(lookup_model).filter_by(**kwargs).first()
                if res:
                    mapped_row[field_map[key]] = res.id
                else:
                    mapped_row[field_map[key]] = None
            else:
                mapped_row[field_map[key]] = value
        instance = model()
        for key, value in mapped_row.items():
            setattr(instance, key, value)
        if 'record_type' in mapping:
            instance.record_type = mapping['record_type']
        self.session.add(instance)

    def _create_tables(self):
        for name, mapping in self.mappings.items():
            self._create_table(mapping)
        self.metadata.create_all()

    def _fields_for_mapping(self, mapping):
        fields = []
        for sf_field, db_field in mapping.get('fields', {}).items():
            fields.append({ 'sf': sf_field, 'db': db_field })
        for sf_field, lookup in mapping.get('lookups', {}).items():
            fields.append({ 'sf': sf_field, 'db': lookup['key_field'] })
        return fields

    def _create_table(self, mapping):
        model_name = '{}Model'.format(mapping['table'])
        mapper_kwargs = {}
        table_kwargs = {}
        if mapping['table'] in self.models:
            mapper_kwargs['non_primary'] = True
            table_kwargs['extend_existing'] = True
        else:
            self.models[mapping['table']] = type(model_name, (object,), {})
        
        fields = []
        fields.append(Column('id', Integer, primary_key=True))
        if 'record_type' in mapping:
            fields.append(Column('record_type', Unicode(255)))
        for field in self._fields_for_mapping(mapping):
            fields.append(Column(field['db'], Unicode(255)))
        t = Table(
            mapping['table'],
            self.metadata,
            *fields,
            **table_kwargs
        )
        
        
        mapper(self.models[mapping['table']], t, **mapper_kwargs)



from cumulusci.tasks.robotframework.robotframework import Robot
from cumulusci.tasks.robotframework.robotframework import RobotLibDoc
from cumulusci.tasks.robotframework.robotframework import RobotTestDoc



import unittest

from mock import MagicMock
from mock import patch
import responses

from cumulusci.core.config import BaseGlobalConfig
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config import ConnectedAppOAuthConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


@patch('cumulusci.tasks.salesforce.BaseSalesforceTask._update_credentials',
    MagicMock(return_value=None))
class TestSalesforceToolingTask(unittest.TestCase):

    def setUp(self):
        self.api_version = 36.0
        self.global_config = BaseGlobalConfig(
            {'project': {'package': {'api_version': self.api_version}}})
        self.project_config = BaseProjectConfig(self.global_config)
        self.project_config.config['project'] = {
            'package': {
                'api_version': self.api_version,
            }
        }
        self.task_config = TaskConfig()
        self.org_config = OrgConfig({
            'instance_url': 'example.com',
            'access_token': 'abc123',
        }, 'test')
        self.base_tooling_url = 'https://{}/services/data/v{}/tooling/'.format(
            self.org_config.instance_url, self.api_version)

    def test_get_tooling_object(self):
        task = BaseSalesforceApiTask(
            self.project_config, self.task_config, self.org_config)
        obj = task._get_tooling_object('TestObject')
        url = self.base_tooling_url + 'sobjects/TestObject/'
        self.assertEqual(obj.base_url, url)



import logging
import re
import time
from robot.libraries.BuiltIn import BuiltIn
from selenium.common.exceptions import ElementNotInteractableException
from selenium.common.exceptions import StaleElementReferenceException
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.keys import Keys
from SeleniumLibrary.errors import ElementNotFound
from simple_salesforce import SalesforceMalformedRequest
from simple_salesforce import SalesforceResourceNotFound
from cumulusci.robotframework.locators import lex_locators

OID_REGEX = r'[a-zA-Z0-9]{15,18}'

class Salesforce(object):
    ROBOT_LIBRARY_SCOPE = 'GLOBAL'

    def __init__(self, debug=False):
        self.debug = debug
        self.current_page = None
        self._session_records = []
        # Turn off info logging of all http requests 
        logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(logging.WARN)

    @property
    def builtin(self):
        return BuiltIn()   

    @property
    def cumulusci(self):
        return self.builtin.get_library_instance('cumulusci.robotframework.CumulusCI')
 
    @property
    def selenium(self):
        return self.builtin.get_library_instance('SeleniumLibrary')

    def current_app_should_be(self, app_name):
        """ EXPERIMENTAL!!! """
        locator = lex_locators['app_launcher']['current_app'].format(app_name)
        self.selenium.set_focus_to_element(locator)
        elem = self.selenium.get_webelement(locator)
        assert app_name == elem.text

    def _call_selenium(self, method_name, retry, *args, **kwargs):
        """ A wrapper that catches common exceptions and handles them """
        exception = None
        retry_call = False

        # If at a new url, call self._handle_current_page() to inject JS into
        # the page to handle Lightning events
        current_page = self.selenium.get_location()
        if current_page != self.current_page:
            self.current_page = current_page
            self._handle_page_load()

        result = None
        try:
            self._wait_until_loading_is_complete()
            method = getattr(self, method_name)
            result = method(*args, **kwargs)
        except ElementNotFound as e:
            # Retry once if we fail to find an element
            if retry is True:
                retry_call = True
            else:
                exception = e
        except StaleElementReferenceException as e:
            # Retry once if we encounter a stale element
            if retry is True:
                retry_call = True
            else:
                exception = e
        except WebDriverException as e:
            if 'Other element would receive the click' in e.message:
                if retry is True:
                    retry_call = True
                else:
                    exception = e
            else:
                exception = e
        except ElementNotInteractableException as e:
            if retry is True:
                retry_call = True
            else:
                exception = e
        except Exception as e:
            exception = e

        if exception:
            self.selenium.capture_page_screenshot()
            if self.debug:
                self.selenium.log_source()
                from cumulusci.robotframework.utils import set_pdb_trace
                set_pdb_trace()
                raise e
            else:
                raise e

        if retry_call:
            self.builtin.log('Retrying call to method {}'.format(method_name), level='WARN')
            self._call_selenium(method_name, False, *args, **kwargs)

        return result

    def click_modal_button(self, title):
        locator = lex_locators['modal']['button'].format(title)
        self._call_selenium('_click_modal_button', True, locator)

    def _click_modal_button(self, locator):
        self.selenium.set_focus_to_element(locator)
        button = self.selenium.get_webelement(locator)
        button.click()

    def click_object_button(self, title):
        locator = lex_locators['object']['button'].format(title)
        self._call_selenium('_click_object_button', True, locator)

    def _click_object_button(self, locator):
        self.selenium.wait_until_element_is_visible(locator)
        button = self.selenium.get_webelement(locator)
        button.click()
        self.wait_until_modal_is_open()

    def click_related_list_button(self, heading, button_title):
        locator = lex_locators['record']['related']['button'].format(heading, button_title)
        self._call_selenium('_click_related_list_button', True, locator)

    def _click_related_list_button(self, locator):
        self.selenium.set_focus_to_element(locator)
        self.selenium.get_webelement(locator).click()
        self.wait_until_modal_is_open()

    def delete_session_records(self):
        self._session_records.reverse()
        self.builtin.log('Deleting {} records'.format(len(self._session_records)))
        for record in self._session_records:
            self.builtin.log('  Deleting {type} {id}'.format(**record))
            try:
                self.salesforce_delete(record['type'], record['id'])
            except SalesforceResourceNotFound:
                self.builtin.log('    {type} {id} is already deleted'.format(**record))
            except SalesforceMalformedRequest as e:
                self.builtin.log('    {type} {id} could not be deleted:'.format(**record), level='WARN')
                self.builtin.log('      {}'.format(e), level='WARN')
        
    def get_current_record_id(self):
        """ Parses the current url to get the object id of the current record.
            Expects url format like: [a-zA-Z0-9]{15,18}
        """
        url = self.selenium.get_location()
        for part in url.split('/'):
            if re.match(OID_REGEX, part):
                return part
        raise AssertionError("Could not parse record id from url: {}".format(url))

    def get_locator(self, path, *args, **kwargs):
        """ Returns a rendered locator string from the Salesforce lex_locators
            dictionary.  This can be useful if you want to use an element in
            a different way than the built in keywords allow.
        """
        locator = lex_locators
        for key in path.split('.'):
            locator = locator[key]
        return locator.format(*args, **kwargs)

    def get_record_type_id(self, obj_type, developer_name):
        soql = "SELECT Id FROM RecordType WHERE SObjectType='{}' and DeveloperName='{}'".format(
            obj_type,
            developer_name,
        )
        res = self.cumulusci.sf.query_all(soql)
        return res['records'][0]['Id']

    def get_related_list_count(self, heading):
        locator = lex_locators['record']['related']['count'].format(heading)
        return self._call_selenium('_get_related_list_count', True, locator)

    def _get_related_list_count(self, locator):
        count = self.selenium.get_webelement(locator).text
        count = count.replace('(','').replace(')','')
        return int(count)
        
    def go_to_object_home(self, obj_name):
        """ Navigates to the Home view of a Salesforce Object """
        url = self.cumulusci.org.lightning_base_url
        url = '{}/one/one.app#/sObject/{}/home'.format(url, obj_name)
        self.selenium.go_to(url)
        time.sleep(5)
        self._wait_until_loading_is_complete()
    
    def go_to_object_list(self, obj_name, filter_name=None):
        """ Navigates to the Home view of a Salesforce Object """
        url = self.cumulusci.org.lightning_base_url
        url = '{}/one/one.app#/sObject/{}/list'.format(url, obj_name)
        if filter_name:
            url += '?filterName={}'.format(filter_name)
        self.selenium.go_to(url)
        time.sleep(5)
        self._wait_until_loading_is_complete()

    def go_to_record_home(self, obj_id, filter_name=None):
        """ Navigates to the Home view of a Salesforce Object """
        url = self.cumulusci.org.lightning_base_url
        url = '{}/one/one.app#/sObject/{}/view'.format(url, obj_id)
        self.selenium.go_to(url)
        time.sleep(5)
        self._wait_until_loading_is_complete()

    def go_to_setup_home(self):
        """ Navigates to the Home tab of Salesforce Setup """
        url = self.cumulusci.org.lightning_base_url
        self.selenium.go_to(url + '/one/one.app#/setup/SetupOneHome/home')
        time.sleep(5)
        self._wait_until_loading_is_complete()

    def go_to_setup_object_manager(self):
        """ Navigates to the Object Manager tab of Salesforce Setup """
        url = self.cumulusci.org.lightning_base_url
        self.selenium.go_to(url + '/one/one.app#/setup/ObjectManager/home')
        time.sleep(5)
        self._wait_until_loading_is_complete()

    def header_field_should_have_value(self, label):
        """ Validates that a field in the record header has a text value.
            NOTE: Use other keywords for non-string value types
        """
        locator = lex_locators['record']['header']['field_value'].format(label)
        self._call_selenium('_header_field_value_should_exist', True, locator)

    def header_field_should_not_have_value(self, label):
        """ Validates that a field in the record header does not have a value.
            NOTE: Use other keywords for non-string value types
        """
        locator = lex_locators['record']['header']['field_value'].format(label)
        self._call_selenium('_header_field_value_should_not_exist', True, locator)

    def header_field_should_have_link(self, label):
        """ Validates that a field in the record header has a link as its value """
        locator = lex_locators['record']['header']['field_value_link'].format(label)
        self._call_selenium('_header_field_value_should_exist', True, locator)
    
    def header_field_should_not_have_link(self, label):
        """ Validates that a field in the record header does not have a link as its value """
        locator = lex_locators['record']['header']['field_value_link'].format(label)
        self._call_selenium('_header_field_value_should_not_exist', True, locator)

    def header_field_should_be_checked(self, label):
        """ Validates that a checkbox field in the record header is checked """
        locator = lex_locators['record']['header']['field_value_checked'].format(label)
        self._call_selenium('_header_field_value_should_exist', True, locator)

    def header_field_should_be_unchecked(self, label):
        """ Validates that a checkbox field in the record header is unchecked """
        locator = lex_locators['record']['header']['field_value_unchecked'].format(label)
        self._call_selenium('_header_field_value_should_exist', True, locator)

    def _header_field_value_should_exist(self, locator):
        self.selenium.page_should_contain_element(locator)
    
    def _header_field_value_should_not_exist(self, locator):
        self.selenium.page_should_not_contain_element(locator)

    def open_app_launcher(self):
        """ EXPERIMENTAL!!! """
        locator = lex_locators['app_launcher']['button']
        self._call_selenium('_open_app_launcher', True, locator)

    def _open_app_launcher(self, locator):
        self.builtin.log('Hovering over App Launcher button')
        self.selenium.mouse_over(locator)
        self.builtin.log('Clicking App Launcher button')
        self.selenium.get_webelement(locator).click()
        self.builtin.log('Waiting for modal to open')
        self.wait_until_modal_is_open()

    def populate_field(self, name, value):
        locator = lex_locators['object']['field'].format(name)
        self._call_selenium('_populate_field', True, locator, value)

    def populate_lookup_field(self, name, value):
        self.populate_field(name, value)
        locator = lex_locators['object']['field_lookup_value'].format(value)
        self._call_selenium('_populate_lookup_field', True, locator)
    
    def _populate_lookup_field(self, locator):
        self.selenium.set_focus_to_element(locator)
        self.selenium.get_webelement(locator).click()
        time.sleep(.5)

    def _populate_field(self, locator, value):
        self.selenium.set_focus_to_element(locator)
        field = self.selenium.get_webelement(locator)
        field.clear()
        time.sleep(.5)
        field.send_keys(value)
        time.sleep(.5)

    def populate_form(self, **kwargs):
        for name, value in kwargs.items():
            time.sleep(.5)
            locator = lex_locators['object']['field'].format(name)
            self._call_selenium('_populate_field', True, locator, value)

    def remove_session_record(self, obj_type, obj_id):
        try:
            self._session_records.remove({
                'type': obj_type,
                'id': obj_id,
            })
        except ValueError:
            self.builtin.log('Did not find record {} {} in the session records list'.format(obj_type, obj_id))

    def select_record_type(self, label):
        self.wait_until_modal_is_open()
        locator = lex_locators['object']['record_type_option'].format(label)
        self._call_selenium('_select_record_type', True, locator)

    def _select_record_type(self, locator):
        self.selenium.get_webelement(locator).click()
        time.sleep(1)
        locator = lex_locators['modal']['button'].format('Next')
        self.selenium.click_button('Next')
        time.sleep(5)

    def select_app_launcher_app(self, app_name):
        """ EXPERIMENTAL!!! """
        locator = lex_locators['app_launcher']['app_link'].format(app_name)
        self.builtin.log('Opening the App Launcher')
        self.open_app_launcher()
        self._call_selenium('_select_app_launcher_app', True, locator)

    def _select_app_launcher_app(self, locator):
        self.builtin.log('Getting the web element for the app')
        self.selenium.set_focus_to_element(locator)
        elem = self.selenium.get_webelement(locator)
        self.builtin.log('Getting the parent link from the web element')
        link = elem.find_element_by_xpath('../../..')
        self.selenium.set_focus_to_element(link)
        self.builtin.log('Clicking the link')
        link.click()
        self.builtin.log('Waiting for modal to close')
        self.wait_until_modal_is_closed()

    def select_app_launcher_tab(self, tab_name):
        """ EXPERIMENTAL!!! """
        locator = lex_locators['app_launcher']['tab_link'].format(tab_name)
        self.builtin.log('Opening the App Launcher')
        self.open_app_launcher()
        self._call_selenium('_select_app_launcher_tab', True, locator)

    def _select_app_launcher_tab(self, locator):
        self.builtin.log('Clicking App Tab')
        self.selenium.set_focus_to_element(locator)
        self.selenium.get_webelement(locator).click()
        self.builtin.log('Waiting for modal to close')
        self.wait_until_modal_is_closed()

    def salesforce_delete(self, obj_name, obj_id):
        """ Deletes a Saleforce object by id and returns the dict result """
        self.builtin.log('Deleting {} with Id {}'.format(obj_name, obj_id))
        obj_class = getattr(self.cumulusci.sf, obj_name)
        res = obj_class.delete(obj_id)
        self.remove_session_record(obj_name, obj_id)

    def salesforce_get(self, obj_name, obj_id):
        """ Gets a Salesforce object by id and returns the dict result """
        self.builtin.log('Getting {} with Id {}'.format(obj_name, obj_id))
        obj_class = getattr(self.cumulusci.sf, obj_name)
        return obj_class.get(obj_id)

    def salesforce_insert(self, obj_name, **kwargs):
        """ Inserts a Salesforce object setting fields using kwargs and returns the id """
        self.builtin.log('Inserting {} with values {}'.format(obj_name, kwargs))
        obj_class = getattr(self.cumulusci.sf, obj_name)
        res = obj_class.create(kwargs)
        self.store_session_record(obj_name, res['id'])
        return res['id']

    def salesforce_query(self, obj_name, **kwargs):
        """ Constructs and runs a simple SOQL query and returns the dict results """
        query = 'SELECT '
        if 'select' in kwargs:
            query += kwargs['select']
        else:
            query += 'Id'
        query += ' FROM {}'.format(obj_name)
        where = []
        for key, value in kwargs.items():
            if key == 'select':
                continue
            where.append("{} = '{}'".format(key, value))
        if where:
            query += ' WHERE ' + ' AND '.join(where)
        self.builtin.log('Running SOQL Query: {}'.format(query))
        return self.cumulusci.sf.query_all(query).get('records', [])

    def salesforce_update(self, obj_name, obj_id, **kwargs):
        """ Updates a Salesforce object by id and returns the dict results """
        self.builtin.log('Updating {} {} with values {}'.format(obj_name, obj_id, kwargs))
        obj_class = getattr(self.cumulusci.sf, obj_name)
        return obj_class.update(obj_id, kwargs)
        
    def soql_query(self, query):
        """ Runs a simple SOQL query and returns the dict results """
        self.builtin.log('Running SOQL Query: {}'.format(query))
        return self.cumulusci.sf.query_all(query)

    def store_session_record(self, obj_type, obj_id):
        """ Stores a Salesforce record's id for use in the Delete Session Records keyword """
        self.builtin.log('Storing {} {} to session records'.format(obj_type, obj_id))
        self._session_records.append({
            'type': obj_type,
            'id': obj_id,
        })

    def wait_until_modal_is_open(self):
        """ EXPERIMENTAL!!! """
        self._call_selenium('_wait_until_modal_is_open', True)

    def _wait_until_modal_is_open(self):
        self.selenium.wait_until_element_is_not_visible(
            lex_locators['spinner'],
            timeout=15,
        )
        self.selenium.wait_until_element_is_visible(
            lex_locators['modal']['is_open'],
            timeout=15,
        )
        time.sleep(3)

    def wait_until_modal_is_closed(self):
        """ EXPERIMENTAL!!! """
        self._call_selenium('_wait_until_modal_is_closed', True)

    def _wait_until_modal_is_closed(self):
        self.selenium.wait_until_element_is_not_visible(
            lex_locators['spinner'],
            timeout=15,
        )
        self.selenium.wait_until_element_is_not_visible(
            lex_locators['modal']['is_open'],
            timeout=15,
        )
        time.sleep(3)

    def wait_until_loading_is_complete(self):
        """ EXPERIMENTAL!!! """
        self._call_selenium('_wait_until_loading_is_complete', True)

    def _wait_until_loading_is_complete(self):
        self.selenium.wait_until_element_is_not_visible(
            lex_locators['loading_box'],
        )
        self.selenium.wait_until_element_is_not_visible(
            lex_locators['spinner'],
        )
        self.selenium.wait_until_page_contains_element(
            lex_locators['desktop_rendered'],
        )

    def _handle_page_load(self):
        """ EXPERIMENTAL!!! """
        # Bypass this method for now and just return.  This is here as a prototype
        return
        self._wait_until_loading_is_complete()
        self.selenium.execute_javascript("""
            function cumulusciDoneRenderingHandler(e) {
                elements = $A.getComponent(e.source.getGlobalId()).getElements()
                if (elements.length > 0) {
                    elements[0].classList.add('cumulusci-done-rendering');
                }
            }
            $A.getRoot().addEventHandler('aura:doneRendering', cumulusciDoneRenderingHandler);
            """
        )
            



""" CumulusCI Tasks for running Apex Tests """

import io
import cgi
import json
import httplib

from simple_salesforce import SalesforceGeneralError

from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.core.exceptions import TaskOptionsError, ApexTestException
from cumulusci.core.utils import process_bool_arg, decode_to_unicode

APEX_LIMITS = {
    'Soql': {'Label': 'TESTING_LIMITS: Number of SOQL queries', 'SYNC': 100, 'ASYNC': 200},
    'Email': {'Label': 'TESTING_LIMITS: Number of Email Invocations', 'SYNC': 10, 'ASYNC': 10},
    'AsyncCalls': {'Label': 'TESTING_LIMITS: Number of future calls', 'SYNC': 50, 'ASYNC': 50},
    'DmlRows': {'Label': 'TESTING_LIMITS: Number of DML rows', 'SYNC': 10000, 'ASYNC': 10000},
    'Cpu': {'Label': 'TESTING_LIMITS: Maximum CPU time', 'SYNC': 10000, 'ASYNC': 60000},
    'QueryRows': {'Label': 'TESTING_LIMITS: Number of query rows', 'SYNC': 50000, 'ASYNC': 50000},
    'Dml': {'Label': 'TESTING_LIMITS: Number of DML statements', 'SYNC': 150, 'ASYNC': 150},
    'MobilePush': {'Label': 'TESTING_LIMITS: Number of Mobile Apex push calls', 'SYNC': 10,
                   'ASYNC': 10},
    'Sosl': {'Label': 'TESTING_LIMITS: Number of SOSL queries', 'SYNC': 20, 'ASYNC': 20},
    'Callouts': {'Label': 'TESTING_LIMITS: Number of callouts', 'SYNC': 100, 'ASYNC': 100},
}


TEST_RESULT_QUERY = """
SELECT Id,ApexClassId,TestTimestamp,
       Message,MethodName,Outcome,
       RunTime,StackTrace,
       (SELECT 
          Id,Callouts,AsyncCalls,DmlRows,Email,
          LimitContext,LimitExceptions,MobilePush,
          QueryRows,Sosl,Cpu,Dml,Soql 
        FROM ApexTestResults) 
FROM ApexTestResult 
WHERE AsyncApexJobId='{}'
"""


class RunApexTests(BaseSalesforceApiTask):
    """ Task to run Apex tests with the Tooling API and report results """
    api_version = '38.0'
    name = 'RunApexTests'
    task_options = {
        'test_name_match': {
            'description': ('Query to find Apex test classes to run ' +
                            '("%" is wildcard).  Defaults to ' +
                            'project__test__name_match'),
            'required': True,
        },
        'test_name_exclude': {
            'description': ('Query to find Apex test classes to exclude ' +
                            '("%" is wildcard).  Defaults to ' +
                            'project__test__name_exclude'),
        },
        'namespace': {
            'description': ('Salesforce project namespace.  Defaults to ' +
                            'project__package__namespace'),
        },
        'managed': {
            'description': ('If True, search for tests in the namespace ' +
                            'only.  Defaults to False'),
        },
        'poll_interval': {
            'description': ('Seconds to wait between polling for Apex test ' +
                            'results.  Defaults to 3'),
        },
        'retries': {
            'description': 'Number of retries (default=10)',
        },
        'retry_interval': {
            'description': 'Number of seconds to wait before the next retry (default=5),'
        },
        'retry_interval_add': {
            'description': 'Number of seconds to add before each retry (default=5),'
        },
        'junit_output': {
            'description': 'File name for JUnit output.  Defaults to test_results.xml',
        },
        'json_output': {
            'description': 'File name for json output.  Defaults to test_results.json',
        },
    }

    def _init_options(self, kwargs):
        super(RunApexTests, self)._init_options(kwargs)

        self.options['test_name_match'] = self.options.get(
            'test_name_match', self.project_config.project__test__name_match)

        self.options['test_name_exclude'] = self.options.get(
            'test_name_exclude', self.project_config.project__test__name_exclude)

        if self.options['test_name_exclude'] is None:
            self.options['test_name_exclude'] = ''

        self.options['namespace'] = self.options.get(
            'namespace', self.project_config.project__package__namespace)

        self.options['retries'] = self.options.get('retries', 10)

        self.options['retry_interval'] = self.options.get('retry_interval', 5)

        self.options['retry_interval_add'] = self.options.get(
            'retry_interval_add', 5)

        self.options['junit_output'] = self.options.get(
            'junit_output', 'test_results.xml')

        self.options['json_output'] = self.options.get(
            'json_output', 'test_results.json')

        self.options['managed'] = process_bool_arg(
            self.options.get('managed', False))

        self.counts = {}

    # pylint: disable=W0201
    def _init_class(self):
        self.classes_by_id = {}
        self.classes_by_name = {}
        self.job_id = None
        self.results_by_class_name = {}
        self.result = None

    def _get_test_classes(self):
        if self.options['managed']:
            namespace = self.options.get('namespace')
            if not namespace:
                raise TaskOptionsError(
                    'Running tests in managed mode but no namespace available.'
                )
            namespace = "'{}'".format(namespace)
        else:
            namespace = 'null'
        # Split by commas to allow multiple class name matching options
        test_name_match = self.options['test_name_match']
        included_tests = []
        for pattern in test_name_match.split(','):
            if pattern:
                included_tests.append("Name LIKE '{}'".format(pattern))
        # Add any excludes to the where clause
        test_name_exclude = self.options.get('test_name_exclude', '')
        excluded_tests = []
        for pattern in test_name_exclude.split(','):
            if pattern:
                excluded_tests.append("(NOT Name LIKE '{}')".format(pattern))
        # Get all test classes for namespace
        query = ('SELECT Id, Name FROM ApexClass ' +
                 'WHERE NamespacePrefix = {}'.format(namespace))
        if included_tests:
            query += ' AND ({})'.format(' OR '.join(included_tests))
        if excluded_tests:
            query += ' AND {}'.format(' AND '.join(excluded_tests))
        # Run the query
        self.logger.info('Running query: {}'.format(query))
        result = self.tooling.query_all(query)
        self.logger.info('Found {} test classes'.format(result['totalSize']))
        return result

    def _get_test_results(self):
        result = self.tooling.query_all(TEST_RESULT_QUERY.format(self.job_id))
        self.counts = {
            'Pass': 0,
            'Fail': 0,
            'CompileFail': 0,
            'Skip': 0,
        }
        for test_result in result['records']:
            class_name = self.classes_by_id[test_result['ApexClassId']]
            self.results_by_class_name[class_name][test_result[
                'MethodName']] = test_result
            self.counts[test_result['Outcome']] += 1
        test_results = []
        class_names = self.results_by_class_name.keys()
        class_names.sort()
        for class_name in class_names:
            message = 'Class: {}'.format(class_name)
            self.logger.info(message)
            method_names = self.results_by_class_name[class_name].keys()
            method_names.sort()
            for method_name in method_names:
                result = self.results_by_class_name[class_name][method_name]
                message = '\t{}: {}'.format(result['Outcome'],
                                            result['MethodName'])
                duration = result['RunTime']
                result['stats'] = self._get_stats_from_result(result)
                if duration:
                    message += ' ({}s)'.format(duration)
                self.logger.info(message)
                test_results.append({
                    'Children': result.get('children', None),
                    'ClassName': decode_to_unicode(class_name),
                    'Method': decode_to_unicode(result['MethodName']),
                    'Message': decode_to_unicode(result['Message']),
                    'Outcome': decode_to_unicode(result['Outcome']),
                    'StackTrace': decode_to_unicode(
                        result['StackTrace']),
                    'Stats': result.get('stats', None),
                    'TestTimestamp': result.get('TestTimestamp', None),
                })
                if result['Outcome'] in ['Fail', 'CompileFail']:
                    self.logger.info('\tMessage: {}'.format(result['Message']))
                    self.logger.info('\tStackTrace: {}'.format(
                        result['StackTrace']))
        self.logger.info('-' * 80)
        self.logger.info('Pass: {}  Fail: {}  CompileFail: {}  Skip: {}'
                         .format(
                             self.counts['Pass'],
                             self.counts['Fail'],
                             self.counts['CompileFail'],
                             self.counts['Skip'],
                         ))
        self.logger.info('-' * 80)
        if self.counts['Fail'] or self.counts['CompileFail']:
            self.logger.error('-' * 80)
            self.logger.error('Failing Tests')
            self.logger.error('-' * 80)
            counter = 0
            for result in test_results:
                if result['Outcome'] not in ['Fail', 'CompileFail']:
                    continue
                counter += 1
                self.logger.error('{}: {}.{} - {}'.format(counter,
                                                          result['ClassName'],
                                                          result['Method'],
                                                          result['Outcome']))
                self.logger.error('\tMessage: {}'.format(result['Message']))
                self.logger.error('\tStackTrace: {}'.format(
                    result['StackTrace']))
        return test_results

    def _get_stats_from_result(self, result):
        stats = {'duration': result['RunTime']}

        if result['ApexTestResults']:
            for limit_name, details in APEX_LIMITS.items():
                limit_use = result['ApexTestResults']['records'][0][limit_name]
                limit_allowed = details[result['ApexTestResults']
                                        ['records'][0]['LimitContext']]
                stats[details['Label']] = {
                    'used': limit_use, 'allowed': limit_allowed}

        return stats

    def _run_task(self):
        result = self._get_test_classes()
        if result['totalSize'] == 0:
            return
        for test_class in result['records']:
            self.classes_by_id[test_class['Id']] = test_class['Name']
            self.classes_by_name[test_class['Name']] = test_class['Id']
            self.results_by_class_name[test_class['Name']] = {}
        self.logger.info('Queuing tests for execution...')
        ids = self.classes_by_id.keys()
        result = self.tooling._call_salesforce(
            method='POST',
            url=self.tooling.base_url + 'runTestsAsynchronous',
            json={'classids': ','.join(str(id) for id in ids)},
        )
        if result.status_code != httplib.OK:

            raise SalesforceGeneralError(result.url,
                                         result.path,
                                         result.status_code,
                                         result.content)
        self.job_id = result.json()
        self._wait_for_tests()
        test_results = self._get_test_results()
        self._write_output(test_results)
        if self.counts.get('Fail') or self.counts.get('CompileFail'):
            raise ApexTestException(
                '{} tests failed and {} tests failed compilation'.format(
                    self.counts.get('Fail'), self.counts.get('CompileFail')
                )
            )

    def _wait_for_tests(self):
        self._poll_interval_s = int(self.options.get('poll_interval', 1))
        self._poll()

    def _poll_action(self):
        self._retry()
        counts = {
            'Aborted': 0,
            'Completed': 0,
            'Failed': 0,
            'Holding': 0,
            'Preparing': 0,
            'Processing': 0,
            'Queued': 0,
        }
        for test_queue_item in self.result['records']:
            counts[test_queue_item['Status']] += 1
        self.logger.info('Completed: {}  Processing: {}  Queued: {}'.format(
            counts['Completed'],
            counts['Processing'],
            counts['Queued'],
        ))
        if counts['Queued'] == 0 and counts['Processing'] == 0:
            self.logger.info('Apex tests completed')
            self.poll_complete = True

    def _try(self):
        self.result = self.tooling.query_all(
            "SELECT Id, Status, ApexClassId FROM ApexTestQueueItem " +
            "WHERE ParentJobId = '{}'".format(self.job_id))

    def _write_output(self, test_results):
        junit_output = self.options['junit_output']
        with io.open(junit_output, mode='w', encoding='utf-8') as f:
            f.write(u'<testsuite tests="{}">\n'.format(len(test_results)))
            for result in test_results:
                s = '  <testcase classname="{}" name="{}"'.format(
                    result['ClassName'], result['Method'])
                if ('Stats' in result and result['Stats']
                        and 'duration' in result['Stats']):
                    s += ' time="{}"'.format(result['Stats']['duration'])
                if result['Outcome'] in ['Fail', 'CompileFail']:
                    s += '>\n'
                    s += '    <failure type="failed" ' 
                    if result['Message']:      
                        s += 'message="{}">'.format(cgi.escape(result['Message']))
                    else:
                        s += '>'
                    
                    if result['StackTrace']:
                        s += '<![CDATA[{}]]>'.format(cgi.escape(result['StackTrace']))
                    s += '</failure>\n'
                    s += '  </testcase>\n'
                else:
                    s += ' />\n'
                f.write(unicode(s))
            f.write(u'</testsuite>')

        json_output = self.options['json_output']
        with io.open(json_output, mode='w', encoding='utf-8') as f:
            f.write(unicode(json.dumps(test_results, indent=4)))



import glob
import os
import lxml.etree as ET
from xml.sax.saxutils import escape
from cumulusci.core.tasks import BaseTask

class RemoveElementsXPath(BaseTask):
    task_options = {
        'elements': {
            'description': 'A list of dictionaries containing path and xpath keys. The path key is a file path that supports wildcards and xpath is the xpath for the elements to remove.  Multiple dictionaries can be passed in the list to run multiple removal queries in the same task.  Metadata elements in the xpath need to be prefixed with ns:, for example: ./ns:Layout/ns:relatedLists',
            'required': True,
        },
        'chdir': {
            'description': 'Change the current directory before running the replace',
        }
    }

    def _run_task(self):
        cwd = os.getcwd()
        chdir = self.options.get('chdir')
        if chdir:
            self.logger.info('Changing directory to {}'.format(chdir))
            os.chdir(chdir)
        for element in self.options['elements']:
            self._process_element(element)
        if chdir:
            os.chdir(cwd)
   
    def _process_element(self, element):
        self.logger.info(
            'Removing elements matching {xpath} from {path}'.format(
                **element
            )
        )
        for f in glob.glob(element['path']):
            with open(f, 'rw') as fp:
                orig = fp.read()
                fp.seek(0)
                root = ET.parse(open(f))
                res = root.findall(element['xpath'].replace('ns:','{http://soap.sforce.com/2006/04/metadata}'))
                for element in res:
                    element.getparent().remove(element)
                processed = '{}\n{}\n'.format(
                    '<?xml version="1.0" encoding="UTF-8"?>',
                    ET.tostring(root),
                )
                if orig != processed:
                    self.logger.info('Modified {}'.format(f))
                    fp = open(f, 'w')
                    fp.write(processed)



import unittest

from mock import MagicMock
from mock import patch
import responses

from cumulusci.core.config import BaseGlobalConfig
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config import ConnectedAppOAuthConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.tasks.apex.testrunner import RunApexTests

@patch('cumulusci.tasks.salesforce.BaseSalesforceTask._update_credentials',
    MagicMock(return_value=None))
class TestRunApexTests(unittest.TestCase):

    def setUp(self):
        self.api_version = 38.0
        self.global_config = BaseGlobalConfig(
            {'project': {'api_version': self.api_version}})
        self.task_config = TaskConfig()
        self.task_config.config['options'] = {
            'junit_output': 'results_junit.xml',
            'poll_interval': 1,
            'test_name_match': '%_TEST',
        }
        self.project_config = BaseProjectConfig(self.global_config)
        self.project_config.config['project'] = {'package': {
            'api_version': self.api_version}}
        keychain = BaseProjectKeychain(self.project_config, '')
        app_config = ConnectedAppOAuthConfig()
        keychain.set_connected_app(app_config)
        self.project_config.set_keychain(keychain)
        self.org_config = OrgConfig({
            'id': 'foo/1',
            'instance_url': 'example.com',
            'access_token': 'abc123',
        }, 'test')
        self.base_tooling_url = 'https://{}/services/data/v{}/tooling/'.format(
            self.org_config.instance_url, self.api_version)

    def _mock_apex_class_query(self):
        url = (self.base_tooling_url + 'query/?q=SELECT+Id%2C+Name+' +
            'FROM+ApexClass+WHERE+NamespacePrefix+%3D+null' +
            '+AND+%28Name+LIKE+%27%25_TEST%27%29')
        expected_response = {
            'done': True,
            'records': [{'Id': 1, 'Name': 'TestClass_TEST'}],
            'totalSize': 1,
        }
        responses.add(responses.GET, url, match_querystring=True,
            json=expected_response)

    def _mock_get_test_results(self):
        url = (self.base_tooling_url + 'query/?q=%0ASELECT+Id%2CApexClassId%2CTestTimestamp%2C%0A+++++++Message%2CMethodName%2COutcome%2C%0A+++++++RunTime%2CStackTrace%2C%0A+++++++%28SELECT+%0A++++++++++Id%2CCallouts%2CAsyncCalls%2CDmlRows%2CEmail%2C%0A++++++++++LimitContext%2CLimitExceptions%2CMobilePush%2C%0A++++++++++QueryRows%2CSosl%2CCpu%2CDml%2CSoql+%0A++++++++FROM+ApexTestResults%29+%0AFROM+ApexTestResult+%0AWHERE+AsyncApexJobId%3D%27JOB_ID1234567%27%0A')

        expected_response = {
            'done': True,

            'records': [{
                "attributes": {
                    "type": "ApexTestResult",
                    "url": "/services/data/v40.0/tooling/sobjects/ApexTestResult/07M41000009gbT3EAI"
                },
                "ApexClass": {
                    "attributes": {
                        "type": "ApexClass",
                        "url": "/services/data/v40.0/tooling/sobjects/ApexClass/01p4100000Fu4Z0AAJ"
                    },
                    "Name": "EP_TaskDependency_TEST"
                },
                "ApexClassId": 1,
                "ApexLogId": 1,
                "TestTimestamp": "2017-07-18T20:36:04.000+0000",
                "Id": "07M41000009gbT3EAI",
                "Message": "Test Passed",
                "MethodName": "TestMethod",
                "Outcome": "Pass",
                "QueueItem": {
                    "attributes": {
                        "type": "ApexTestQueueItem",
                        "url": "/services/data/v40.0/tooling/sobjects/ApexTestQueueItem/70941000000q7VsAAI"
                    },
                    "Status": "Completed",
                    "ExtendedStatus": "(4/4)"
                },
                "RunTime": 1707,
                "StackTrace": "1. ParentFunction\n2. ChildFunction",
                "Status": 'Completed',
                'Name': 'TestClass_TEST',
                "ApexTestResults": {
                    "size": 1,
                    "totalSize": 1,
                    "done": True,
                    "queryLocator": None,
                    "entityTypeName": "ApexTestResultLimits",
                    "records": [{
                        "attributes": {
                            "type": "ApexTestResultLimits",
                            "url": "/services/data/v40.0/tooling/sobjects/ApexTestResultLimits/05n41000002Y7OQAA0"
                        },
                        "Id": "05n41000002Y7OQAA0",
                        "Callouts": 0,
                        "AsyncCalls": 0,
                        "DmlRows": 5,
                        "Email": 0,
                        "LimitContext": "SYNC",
                        "LimitExceptions": None,
                        "MobilePush": 0,
                        "QueryRows": 20,
                        "Sosl": 0,
                        "Cpu": 471,
                        "Dml": 4,
                        "Soql": 5
                    }]
                }
            }]}

        responses.add(responses.GET, url,
                      match_querystring=True, json=expected_response)

    def _mock_tests_complete(self):
        url = (self.base_tooling_url + 'query/?q=SELECT+Id%2C+Status%2C+' +
            'ApexClassId+FROM+ApexTestQueueItem+WHERE+ParentJobId+%3D+%27' +
            'JOB_ID1234567%27')
        expected_response = {
            'done': True,
            'records': [{'Status': 'Completed'}],
        }
        responses.add(responses.GET, url, match_querystring=True,
            json=expected_response)

    def _mock_run_tests(self):
        url = self.base_tooling_url + 'runTestsAsynchronous'
        expected_response = 'JOB_ID1234567'
        responses.add(responses.POST, url, json=expected_response)

    @responses.activate
    def test_run_task(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_tests_complete()
        self._mock_get_test_results()
        task = RunApexTests(
            self.project_config, self.task_config, self.org_config)
        task()
        self.assertEqual(len(responses.calls), 4)



from __future__ import unicode_literals, print_function
from builtins import chr
import base64
import json
import os
import pickle

from Crypto import Random
from Crypto.Cipher import AES

from cumulusci.core.config import BaseConfig
from cumulusci.core.config import ConnectedAppOAuthConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import ScratchOrgConfig
from cumulusci.core.config import ServiceConfig
from cumulusci.core.exceptions import OrgNotFound
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.exceptions import ServiceNotValid
from cumulusci.core.exceptions import KeychainConnectedAppNotFound


class BaseProjectKeychain(BaseConfig):
    encrypted = False

    def __init__(self, project_config, key):
        super(BaseProjectKeychain, self).__init__()
        self.config = {
            'orgs': {},
            'app': None,
            'services': {},
        }
        self.project_config = project_config
        self.key = key
        self._load_keychain()

    def _load_keychain(self):
        self._load_app()
        self._load_orgs()
        self._load_scratch_orgs()
        self._load_services()

    def _load_app(self):
        pass

    def _load_orgs(self):
        pass

    def _load_scratch_orgs(self):
        """ Creates all scratch org configs for the project in the keychain if
            a keychain org doesn't already exist """
        current_orgs = self.list_orgs()
        if not self.project_config.orgs__scratch:
            return
        for config_name in self.project_config.orgs__scratch.keys():
            if config_name in current_orgs:
                # Don't overwrite an existing keychain org
                continue
            self.create_scratch_org(config_name, config_name)

    def _load_services(self):
        pass
            
    def create_scratch_org(self, org_name, config_name, days=None):
        """ Adds/Updates a scratch org config to the keychain from a named config """
        scratch_config = getattr(self.project_config, 'orgs__scratch__{}'.format(config_name))
        if days is not None:
            # Allow override of scratch config's default days
            scratch_config['days'] = days
        else:
            # Use scratch config days or default of 1 day
            scratch_config.setdefault('days', 1)
        scratch_config['scratch'] = True
        scratch_config.setdefault('namespaced', False)
        scratch_config['config_name'] = config_name
        scratch_config['sfdx_alias'] = '{}__{}'.format(
            self.project_config.project__name,
            org_name,
        )
        org_config = ScratchOrgConfig(scratch_config, org_name)
        self.set_org(org_config)

    def change_key(self, key):
        """ re-encrypt stored services, orgs, and the connected_app
        with the new key """

        connected_app = self.get_connected_app()

        services = {}
        for service_name in self.list_services():
            services[service_name] = self.get_service(service_name)

        orgs = {}
        for org_name in self.list_orgs():
            orgs[org_name] = self.get_org(org_name)

        self.key = key
        if connected_app:
            self.set_connected_app(connected_app)

        if orgs:
            for org_name, org_config in list(orgs.items()):
                self.set_org(org_config)

        if services:
            for service_name, service_config in list(services.items()):
                self.set_service(service_name, service_config)

    def set_connected_app(self, app_config, project=False):
        """ store a connected_app configuration """

        self._set_connected_app(app_config, project)
        self._load_app()

    def _set_connected_app(self, app_config, project):
        self.app = app_config

    def get_connected_app(self):
        """ retrieve the connected app configuration """

        return self._get_connected_app()

    def _get_connected_app(self):
        return self.app

    def remove_org(self, name, global_org=None):
        if name in self.orgs.keys():
            self._remove_org(name, global_org)

    def _remove_org(self, name, global_org):
        del self.orgs[name]
        self._load_orgs()

    def set_org(self, org_config, global_org=False):
        if isinstance(org_config, ScratchOrgConfig):
            org_config.config['scratch'] = True
        self._set_org(org_config, global_org)
        self._load_orgs()

    def _set_org(self, org_config, global_org):
        self.orgs[org_config.name] = org_config

    def get_default_org(self):
        """ retrieve the name and configuration of the default org """
        for org in self.list_orgs():
            org_config = self.get_org(org)
            if org_config.default:
                return org, org_config
        return None, None

    def set_default_org(self, name):
        """ set the default org for tasks by name key """
        org = self.get_org(name)
        self.unset_default_org()
        org.config['default'] = True
        self.set_org(org)

    def unset_default_org(self):
        """ unset the default orgs for tasks """
        for org in self.list_orgs():
            org_config = self.get_org(org)
            if org_config.default:
                del org_config.config['default']
                self.set_org(org_config)

    def get_org(self, name):
        """ retrieve an org configuration by name key """
        if name not in self.orgs:
            self._raise_org_not_found(name)
        return self._get_org(name)

    def _get_org(self, name):
        return self.orgs.get(name)

    def _raise_org_not_found(self, name):
        raise OrgNotFound(
            'Org named {} was not found in keychain'.format(name))

    def list_orgs(self):
        """ list the orgs configured in the keychain """
        orgs = list(self.orgs.keys())
        orgs.sort()
        return orgs

    def set_service(self, name, service_config, project=False):
        """ Store a ServiceConfig in the keychain """
        if not self.project_config.services or name not in self.project_config.services:
            self._raise_service_not_valid(name)
        self._validate_service(name, service_config)
        self._set_service(name, service_config, project)
        self._load_services()

    def _set_service(self, name, service_config, project=False):
        self.services[name] = service_config

    def get_service(self, name):
        """ Retrieve a stored ServiceConfig from the keychain or exception

        :param name: the service name to retrieve
        :type name: str

        :rtype ServiceConfig
        :return the configured Service
        """
        if not self.project_config.services or name not in self.project_config.services:
            self._raise_service_not_valid(name)
        if name not in self.services:
            self._raise_service_not_configured(name)

        return self._get_service(name)

    def _get_service(self, name):
        return self.services.get(name)

    def _validate_service(self, name, service_config):
        missing_required = []
        attr_key = 'services__{0}__attributes'.format(name)
        for atr, config in list(getattr(self.project_config, attr_key).items()):
            if config.get('required') is True and not getattr(service_config, atr):
                missing_required.append(atr)

        if missing_required:
            self._raise_service_not_valid(name)

    def _raise_service_not_configured(self, name):
        raise ServiceNotConfigured(
            'Service named {} is not configured for this project. Configured services are: {}'.format(name, ', '.join(list(self.services)))
        )

    def _raise_service_not_valid(self, name):
        raise ServiceNotValid(
            'Service named {} is not valid for this project'.format(name))

    def list_services(self):
        """ list the services configured in the keychain """
        services = list(self.services.keys())
        services.sort()
        return services


class EnvironmentProjectKeychain(BaseProjectKeychain):
    """ A project keychain that stores org credentials in environment variables """
    encrypted = False
    org_var_prefix = 'CUMULUSCI_ORG_'
    app_var = 'CUMULUSCI_CONNECTED_APP'
    service_var_prefix = 'CUMULUSCI_SERVICE_'

    def _get_env(self):
        """ loads the environment variables as unicode if ascii """
        try:
            return [(k.decode(), v.decode()) for k,v in list(os.environ.items())]
        except AttributeError:
            return list(os.environ.items())

    def _load_app(self):
        try:
            app = os.environ.get(self.app_var.encode('ascii'))
        except TypeError:
            app = os.environ.get(self.app_var)
        if app:
            self.app = ConnectedAppOAuthConfig(json.loads(app))

    def _load_orgs(self):
        for key, value in self._get_env():
            if key.startswith(self.org_var_prefix):
                org_config = json.loads(value)
                org_name = key[len(self.org_var_prefix):]
                if org_config.get('scratch'):
                    self.orgs[org_name] = ScratchOrgConfig(json.loads(value), org_name)
                else:
                    self.orgs[org_name] = OrgConfig(json.loads(value), org_name)

    def _load_services(self):
        for key, value in self._get_env():
            if key.startswith(self.service_var_prefix):
                service_config = json.loads(value)
                service_name = key[len(self.service_var_prefix):]
                self._set_service(service_name, ServiceConfig(service_config))



BS = 16
pad = lambda s: s + (BS - len(s) % BS) * chr(BS - len(s) % BS).encode('ascii')
unpad = lambda s: s[0:-ord(s[-1])]


class BaseEncryptedProjectKeychain(BaseProjectKeychain):
    """ Base class for building project keychains that use AES encryption for securing stored org credentials """
    encrypted = True

    def _set_connected_app(self, app_config, project):
        encrypted = self._encrypt_config(app_config)
        self._set_encrypted_connected_app(encrypted, project)

    def _set_encrypted_connected_app(self, encrypted, project):
        self.app = encrypted

    def _get_connected_app(self):
        if self.app:
            return self._decrypt_config(ConnectedAppOAuthConfig, self.app)

    def _get_service(self, name):
        return self._decrypt_config(ServiceConfig, self.services[name])

    def _set_service(self, service, service_config, project):
        encrypted = self._encrypt_config(service_config)
        self._set_encrypted_service(service, encrypted, project)

    def _set_encrypted_service(self, service, encrypted, project):
        self.services[service] = encrypted

    def _set_org(self, org_config, global_org):
        encrypted = self._encrypt_config(org_config)
        self._set_encrypted_org(org_config.name, encrypted, global_org)

    def _set_encrypted_org(self, name, encrypted, global_org):
        self.orgs[name] = encrypted

    def _get_org(self, name):
        return self._decrypt_config(OrgConfig, self.orgs[name], extra=[name])

    def _get_cipher(self, iv=None):
        if iv is None:
            iv = Random.new().read(AES.block_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return cipher, iv

    def _encrypt_config(self, config):
        pickled = pickle.dumps(config.config)
        pickled = pad(pickled)
        #pickled = base64.b64encode(pickled)
        cipher, iv = self._get_cipher()
        encrypted = base64.b64encode(iv + cipher.encrypt(pickled))
        return encrypted

    def _decrypt_config(self, config_class, encrypted_config, extra=None):
        if not encrypted_config:
            if extra:
                return config_class(None, *extra)
            else:
                return config_class()
        encrypted_config = base64.b64decode(encrypted_config)
        iv = encrypted_config[:16]
        cipher, iv = self._get_cipher(iv)
        pickled = cipher.decrypt(encrypted_config[16:])
        config_dict = pickle.loads(pickled)
        args = [config_dict]
        if extra:
            args += extra
        return self._construct_config(config_class, args)

    def _construct_config(self, config_class, args):
        if args[0].get('scratch'):
            config_class = ScratchOrgConfig
        
        return config_class(*args)


class EncryptedFileProjectKeychain(BaseEncryptedProjectKeychain):
    """ An encrypted project keychain that stores in the project's local directory """

    @property
    def config_local_dir(self):
        return os.path.join(
            os.path.expanduser('~'),
            self.project_config.global_config_obj.config_local_dir,
        )

    @property
    def project_local_dir(self):
        return self.project_config.project_local_dir

    def _load_files(self, dirname, extension, key):
        for item in os.listdir(dirname):
            if item.endswith(extension):
                with open(os.path.join(dirname, item), 'r') as f_item:
                    config = f_item.read()
                name = item.replace(extension, '')
                if not key in self.config:
                    self.config[key] = []
                self.config[key][name] = config

    def _load_file(self, dirname, filename, key):
        full_path = os.path.join(dirname, filename)
        if not os.path.exists(full_path):
            return
        with open(os.path.join(dirname, filename), 'r') as f_item:
            config = f_item.read()
        self.config[key] = config

    def _load_app(self):
        self._load_file(self.config_local_dir, 'connected.app', 'app')
        self._load_file(self.project_local_dir, 'connected.app', 'app')

    def _load_orgs(self):
        self._load_files(self.config_local_dir, '.org', 'orgs')
        self._load_files(self.project_local_dir, '.org', 'orgs')

    def _load_services(self):
        self._load_files(self.config_local_dir, '.service', 'services')
        self._load_files(self.project_local_dir, '.service', 'services')

    def _remove_org(self, name, global_org):
        if global_org:
            full_path = os.path.join(self.config_local_dir, '{}.org'.format(name))
        else:
            full_path = os.path.join(self.project_local_dir, '{}.org'.format(name))
        if not os.path.exists(full_path):
            kwargs = {'name': name}
            if not global_org:
                raise OrgNotFound('Could not find org named {name} to delete.  Deleting in project org mode.  Is {name} a global org?'.format(**kwargs))
            raise OrgNotFound('Could not find org named {name} to delete.  Deleting in global org mode.  Is {name} a project org instead of a global org?'.format(**kwargs))
           
        os.remove(full_path) 
        self._load_orgs()

    def _set_encrypted_connected_app(self, encrypted, project):
        if project:
            filename = os.path.join(self.project_local_dir, 'connected.app')
        else:
            filename = os.path.join(self.config_local_dir, 'connected.app')
        with open(filename, 'wb') as f_org:
            f_org.write(encrypted)
        self.app = encrypted

    def _set_encrypted_org(self, name, encrypted, global_org):
        if global_org:
            filename = os.path.join(
                self.config_local_dir, '{}.org'.format(name))
        else:
            filename = os.path.join(
                self.project_local_dir, '{}.org'.format(name))
        with open(filename, 'wb') as f_org:
            f_org.write(encrypted)

    def _set_encrypted_service(self, name, encrypted, project):
        if project:
            filename = os.path.join(
                self.project_local_dir, '{}.service'.format(name))
        else:
            filename = os.path.join(
                self.config_local_dir, '{}.service'.format(name))
        with open(filename, 'wb') as f_service:
            f_service.write(encrypted)

    def _raise_org_not_found(self, name):
        raise OrgNotFound(
            'Org information could not be found.  Expected to find encrypted file at {}/{}.org'.format(
                self.project_local_dir,
                name
            )
        )

    def _raise_service_not_configured(self, name):
        raise ServiceNotConfigured(
            'Service configuration could not be found.  Expected to find encrypted file at {}/{}.org'.format(
                self.project_local_dir,
                name
            )
        )



from cumulusci.core.exceptions import CumulusCIException


class MetadataApiError(CumulusCIException):

    def __init__(self, message, response):
        super(MetadataApiError, self).__init__(message)
        self.response = response

class MetadataComponentFailure(MetadataApiError):
    pass

class MissingOAuthError(CumulusCIException):
    pass


class MissingOrgCredentialsError(CumulusCIException):
    pass



from cumulusci.core.tasks import BaseTask
from cumulusci.tasks.salesforce import BaseSalesforceTask
from robot.run import run
from robot.libdoc import libdoc
from robot.testdoc import testdoc

class Robot(BaseSalesforceTask):
    task_options = {
        'suites': {
            'description': 'Paths to test case files/directories to be executed similarly as when running the robot command on the command line.  Defaults to "tests" to run all tests in the tests directory',
            'required': True,
        },
        'test': {
            'description': 'Run a specific test by name',
        },
        'include': {
            'description': 'Includes tests with a given tag',
        },
        'exclude': {
            'description': 'Excludes tests with a given tag',
        },
        'vars': {
            'description': 'Pass values to override variables in the format VAR1:foo,VAR2:bar',
        },
        'options': {
            'description': 'A dictionary of options to robot.run method.  See docs here for format.  NOTE: There is no cci CLI support for this option since it requires a dictionary.  Use this option in the cumulusci.yml when defining custom tasks where you can easily create a dictionary in yaml.',
        },
    }

    def _init_options(self, kwargs):
        super(Robot, self)._init_options(kwargs)

        # Set default for suites
        if 'suites' not in self.options:
            self.options['suites'] = 'tests'

        # Initialize the vars list and add the org name to it
        if 'vars' in self.options:
            if not isinstance(self.options['vars'], list):
                new_vars = []
                for var in self.options['vars'].split(','):
                    new_vars.append(var.strip())
                self.options['vars'] = new_vars
        else:
            self.options['vars'] = []
        self.options['vars'].append('ORG:{}'.format(self.org_config.name))

        # Initialize options as a dict
        if 'options' not in self.options:
            self.options['options'] = {}

    def _run_task(self):
        options = self.options['options'].copy()
        if 'test' in self.options:
            options['test'] = self.options['test']
        if 'include' in self.options:
            options['include_tag'] = self.options['include']
        if 'exclude' in self.options:
            options['exclude_tag'] = self.options['exclude']

        run(
            self.options['suites'], 
            variable=self.options['vars'],
            **self.options['options']
        )

class RobotLibDoc(BaseTask):
    task_options = {
        'path': {
            'description': "The path to the robot library to be documented.  Can be a python file or a .robot file.",
            'required': True,
        },
        'output': {
            'description': "The output file where the documentation will be written",
            'required': True,
        },
    }

    def _run_task(self):
        return libdoc(
            self.options['path'],
            self.options['output'],
        )

class RobotTestDoc(BaseTask):
    task_options = {
        'path': {
            'description': "The path containing .robot test files",
            'required': True,
        },
        'output': {
            'description': "The output html file where the documentation will be written",
            'required': True,
        },
    }

    def _run_task(self):
        return testdoc(
            self.options['path'],
            self.options['output'],
        )



from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import object
import json
import os
import sys
import webbrowser
import code
import yaml
import time

try:
    import anydbm as dbm
except ImportError:
    import dbm.ndbm as dbm

from contextlib import contextmanager

import click
import pkg_resources
import requests
from plaintable import Table
from rst2ansi import rst2ansi

import cumulusci
from cumulusci.core.config import ConnectedAppOAuthConfig
from cumulusci.core.config import FlowConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import ScratchOrgConfig
from cumulusci.core.config import ServiceConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.config import YamlGlobalConfig
from cumulusci.core.config import YamlProjectConfig
from cumulusci.core.exceptions import ApexTestException
from cumulusci.core.exceptions import BrowserTestFailure
from cumulusci.core.exceptions import ConfigError
from cumulusci.core.exceptions import FlowNotFoundError
from cumulusci.core.exceptions import KeychainConnectedAppNotFound
from cumulusci.core.exceptions import KeychainKeyNotFound
from cumulusci.core.exceptions import OrgNotFound
from cumulusci.salesforce_api.exceptions import MetadataApiError
from cumulusci.salesforce_api.exceptions import MetadataComponentFailure
from cumulusci.core.exceptions import NotInProject
from cumulusci.core.exceptions import ProjectConfigNotFound
from cumulusci.core.exceptions import ScratchOrgException
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.exceptions import TaskNotFoundError
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.exceptions import TaskRequiresSalesforceOrg
from cumulusci.core.utils import import_class
from cumulusci.cli.config import CliConfig
from cumulusci.utils import doc_task
from cumulusci.oauth.salesforce import CaptureSalesforceOAuth
from .logger import init_logger


@contextmanager
def dbm_cache():
    """
    context manager for accessing simple dbm cache
    located at ~/.cumlusci/cache.dbm
    """
    config_dir = os.path.join(
        os.path.expanduser('~'),
        YamlGlobalConfig.config_local_dir,
    )
        
    if not os.path.exists(config_dir):
        os.mkdir(config_dir)

    db = dbm.open(os.path.join(config_dir, 'cache.dbm'), 'c',)
    yield db
    db.close()


def get_installed_version():
    """ returns the version name (e.g. 2.0.0b58) that is installed """
    req = pkg_resources.Requirement.parse('cumulusci')
    dist = pkg_resources.WorkingSet().find(req)
    return pkg_resources.parse_version(dist.version)


def get_latest_version():
    """ return the latest version of cumulusci in pypi, be defensive """
    # use the pypi json api https://wiki.python.org/moin/PyPIJSON
    res = requests.get('https://pypi.python.org/pypi/cumulusci/json', timeout=5).json()
    with dbm_cache() as cache:
        cache['cumulusci-latest-timestamp'] = str(time.time())
    return pkg_resources.parse_version(res['info']['version'])


def check_latest_version():
    """ checks for the latest version of cumulusci from pypi, max once per hour """
    check = True

    with dbm_cache() as cache:
        if 'cumulusci-latest-timestamp' in cache:
            delta = time.time() - float(cache['cumulusci-latest-timestamp'])
            check = delta > 3600

    if check:
        result = get_latest_version() > get_installed_version()
        click.echo('Checking the version!')
        if result:
            click.echo(
                "An update to CumulusCI is available. Use pip install --upgrade cumulusci to update.")

def check_org_expired(config, org_name, org_config):
    if org_config.scratch and org_config.date_created and org_config.expired:
        click.echo(click.style('The scratch org is expired', fg='yellow'))
        if click.confirm('Attempt to recreate the scratch org?', default=True):
            config.keychain.create_scratch_org(
                org_name,
                org_config.config_name,
                org_config.days,
            )
            click.echo('Org config was refreshed, attempting to recreate scratch org')
            org_config = config.keychain.get_org(org_name)
            org_config.create_org()
        else:
            raise click.ClickException('The target scratch org is expired.  You can use cci org remove {} to remove the org and then recreate the config manually'.format(org_name))

    return org_config

def handle_exception_debug(config, debug, e, throw_exception=None, no_prompt=None):
    if debug:
        import pdb
        import traceback
        traceback.print_exc()
        pdb.post_mortem()
    else:
        if throw_exception:
            raise throw_exception
        else:
            handle_sentry_event(config, no_prompt)
            raise

def render_recursive(data, indent=None):
    if indent is None:
        indent = 0
    indent_str = ' ' * indent
    if isinstance(data, list):
        for item in data:
            render_recursive(item, indent=indent+4)
    elif isinstance(data, dict):
        for key, value in data.items():
            key_str = click.style(unicode(key) + ':', bold=True)
            if isinstance(value, list):
                render_recursive(value, indent=indent+4)
            elif isinstance(value, dict):
                click.echo('{}{}'.format(indent_str, key_str))
                render_recursive(value, indent=indent+4)
            else:
                click.echo('{}{} {}'.format(indent_str, key_str, value))

def check_org_overwrite(config, org_name):
    try:
        org = config.keychain.get_org(org_name)
        raise click.ClickException(
            'Org {} already exists.  Use `cci org remove` to delete it.'.format(org_name)
        )
    except OrgNotFound:
        pass
    return True

def make_pass_instance_decorator(obj, ensure=False):
    """Given an object type this creates a decorator that will work
    similar to :func:`pass_obj` but instead of passing the object of the
    current context, it will inject the passed object instance.

    This generates a decorator that works roughly like this::
        from functools import update_wrapper
        def decorator(f):
            @pass_context
            def new_func(ctx, *args, **kwargs):
                return ctx.invoke(f, obj, *args, **kwargs)
            return update_wrapper(new_func, f)
        return decorator
    :param obj: the object instance to pass.
    """
    def decorator(f):
        def new_func(*args, **kwargs):
            ctx = click.get_current_context()
            return ctx.invoke(f, obj, *args[1:], **kwargs)
        return click.decorators.update_wrapper(new_func, f)
    return decorator

try:
    check_latest_version()
except requests.exceptions.RequestException as e:
    click.echo('Error checking cci version:')
    click.echo(e.message) 

try:
    CLI_CONFIG = CliConfig()
except click.UsageError as e:
    click.echo(e.message)
    sys.exit(1)

pass_config = make_pass_instance_decorator(CLI_CONFIG)

def check_connected_app(config):
    check_keychain(config)
    if not config.keychain.get_connected_app():
        raise click.UsageError(
            "Please use the 'org config_connected_app' command to configure the OAuth Connected App to use for this project's keychain")


def check_keychain(config):
    check_project_config(config)
    if config.project_config.keychain and config.project_config.keychain.encrypted and not config.keychain_key:
        raise click.UsageError(
            'You must set the environment variable CUMULUSCI_KEY with the encryption key to be used for storing org credentials')


def check_project_config(config):
    if not config.project_config:
        raise click.UsageError(
            'No project configuration found.  You can use the "project init" command to initilize the project for use with CumulusCI')


def handle_sentry_event(config, no_prompt):
    event = config.project_config.sentry_event
    if not event:
        return

    sentry_config = config.project_config.keychain.get_service('sentry')
    event_url = '{}/{}/{}/?query={}'.format(
        config.project_config.sentry.remote.base_url,
        sentry_config.org_slug,
        sentry_config.project_slug,
        event,
    )
    click.echo(
        'An error event was recorded in sentry.io and can be viewed at the url:\n{}'.format(event_url))

    if not no_prompt and click.confirm(click.style('Do you want to open a browser to view the error in sentry.io?', bold=True)):
        webbrowser.open(event_url)

# Root command


@click.group('main')
@pass_config
def main(config):
    pass


@click.command(name='version', help='Print the current version of CumulusCI')
def version():
    click.echo(cumulusci.__version__)


@click.command(name='shell', help='Drop into a python shell')
@pass_config
@click.pass_context
def shell(ctx, config):
    code.interact(local=dict(globals(), **locals()))

# Top Level Groups


@click.group('project', help="Commands for interacting with project repository configurations")
@pass_config
def project(config):
    pass


@click.group('org', help="Commands for connecting and interacting with Salesforce orgs")
@pass_config
def org(config):
    pass


@click.group('task', help="Commands for finding and running tasks for a project")
@pass_config
def task(config):
    pass


@click.group('flow', help="Commands for finding and running flows for a project")
@pass_config
def flow(config):
    pass


@click.group('service', help="Commands for connecting services to the keychain")
@pass_config
def service(config):
    pass

main.add_command(project)
main.add_command(org)
main.add_command(task)
main.add_command(flow)
main.add_command(version)
main.add_command(shell)
main.add_command(service)

# Commands for group: project


@click.command(name='init',
               help="Initialize a new project for use with the cumulusci toolbelt",
               )
@click.option('--extend', help="If set to the url of another Github repository configured for CumulusCI, creates this project as an extension which depends on the other Github project and all its dependencies")
@pass_config
def project_init(config, extend):
    if not os.path.isdir('.git'):
        raise click.ClickException("You are not in the root of a Git repository")

    if os.path.isfile('cumulusci.yml'):
        raise click.ClickException("This project already has a cumulusci.yml file")

    yml_config = []

    click.echo()
    click.echo(click.style('# Project Info', bold=True, fg='blue'))
    click.echo('The following prompts will collect general information about the project')
    package_config = []

    name = os.path.split(os.getcwd())[-1:][0]
    click.echo()
    click.echo('Enter the project name.  The name is usually the same as your repository name')
    name = click.prompt(click.style('Project Name', bold=True), default=name)
    yml_config.append('project:')
    yml_config.append('    name: {}'.format(name))

    click.echo()
    click.echo("CumulusCI uses an unmanaged package as a container for your project's metadata.  Enter the name of the package you want to use.")
    package_name = click.prompt(click.style('Package Name', bold=True), default=name)
    if package_name and package_name != config.global_config.project__package__name:
        package_config.append('        name: {}'.format(package_name))

    click.echo()
    package_namespace = None
    if click.confirm(click.style("Is this a managed package project?", bold=True), default=False):
        click.echo('Enter the namespace assigned to the managed package for this project')
        package_namespace = click.prompt(click.style('Package Namespace', bold=True), default=name)
        if package_namespace and package_namespace != config.global_config.project__package__namespace:
            package_config.append(
                '        namespace: {}'.format(package_namespace))

    click.echo()
    package_api_version = click.prompt(click.style('Salesforce API Version', bold=True), default=config.global_config.project__package__api_version)
    if package_api_version and package_api_version != config.global_config.project__package__api_version:
        package_config.append(
            '        api_version: {}'.format(package_api_version))

    if package_config:
        yml_config.append('    package:')
        yml_config.extend(package_config)

    if extend:
        yml_config.append('    dependencies:')
        yml_config.append('        - github: {}'.format(extend))

    #     git:
    git_config = []
    click.echo()
    click.echo(click.style('# Git Configuration', bold=True, fg='blue'))
    click.echo('CumulusCI assumes your default branch is master, your feature branches are named feature/*, your beta release tags are named beta/*, and your release tags are release/*.  If you want to use a different branch/tag naming scheme, you can configure the overrides here.  Otherwise, just accept the defaults.')

    git_default_branch = click.prompt(click.style('Default Branch', bold=True), default='master')
    if git_default_branch and git_default_branch != config.global_config.project__git__default_branch:
        git_config.append(
            '        default_branch: {}'.format(git_default_branch))

    git_prefix_feature = click.prompt(click.style('Feature Branch Prefix', bold=True), default='feature/')
    if git_prefix_feature and git_prefix_feature != config.global_config.project__git__prefix_feature:
        git_config.append(
            '        prefix_feature: {}'.format(git_prefix_feature))

    git_prefix_beta = click.prompt(click.style('Beta Tag Prefix', bold=True), default='beta/')
    if git_prefix_beta and git_prefix_beta != config.global_config.project__git__prefix_beta:
        git_config.append('        prefix_beta: {}'.format(git_prefix_beta))

    git_prefix_release = click.prompt(click.style('Release Tag Prefix', bold=True), default='release/')
    if git_prefix_release and git_prefix_release != config.global_config.project__git__prefix_release:
        git_config.append(
            '        prefix_release: {}'.format(git_prefix_release))

    if git_config:
        yml_config.append('    git:')
        yml_config.extend(git_config)

    #     test:
    test_config = []
    click.echo()
    click.echo(click.style('# Apex Tests Configuration', bold=True, fg='blue'))
    click.echo('The CumulusCI Apex test runner uses a SOQL where clause to select which tests to run.  Enter the SOQL pattern to use to match test class names.')
    
    test_name_match = click.prompt(click.style('Test Name Match', bold=True), default=config.global_config.project__test__name_match)
    if test_name_match and test_name_match != config.global_config.project__test__name_match:
        test_config.append('        name_match: {}'.format(test_name_match))
    if test_config:
        yml_config.append('    test:')
        yml_config.extend(test_config)

    if package_namespace:
        yml_config.append('orgs:')
        yml_config.append('    scratch:')
        yml_config.append('        dev_namespaced:')
        yml_config.append('            config_file: orgs/dev.json')
        yml_config.append('            namespaced: True')

    yml_config.append('')

    with open('cumulusci.yml', 'w') as f_yml:
        f_yml.write('\n'.join(yml_config))

    click.echo(click.style("Your project is now initialized for use with CumulusCI", bold=True, fg='green'))
    click.echo(click.style(
        "You can use the project edit command to edit the project's config file", fg='yellow'
    ))

    if os.path.isfile('sfdx-project.json'):
        return

    if not os.path.isdir('src'):
        os.mkdir('src')

    sfdx_project = {
        "packageDirectories": [
            {
                "path": "force-app",
                "default": True,
            }
        ],
        "namespace": package_namespace,
        "sourceApiVersion": package_api_version,
}
    with open('sfdx-project.json','w') as f:
        f.write(json.dumps(sfdx_project))

    if not os.path.isdir('orgs'):
        os.mkdir('orgs')
      
    org_content_url = 'https://raw.githubusercontent.com/SalesforceFoundation/sfdo-package-cookiecutter/master/%7B%7Bcookiecutter.project_name%7D%7D/orgs/{}.json' 
    for org in ['beta', 'dev', 'feature', 'release']:
        with open('orgs/{}.json'.format(org), 'w') as f:
            resp = requests.get(org_content_url.format(org))
            content = resp.content.replace('{{cookiecutter.package_name}}', package_name)
            f.write(content) 
    
        

@click.command(name='info', help="Display information about the current project's configuration")
@pass_config
def project_info(config):
    check_project_config(config)
    click.echo(render_recursive(config.project_config.project))


@click.command(name="dependencies", help="Displays the current dependencies for the project.  If the dependencies section has references to other github repositories, the repositories are inspected and a static list of dependencies is created")
@pass_config
def project_dependencies(config):
    check_project_config(config)
    dependencies = config.project_config.get_static_dependencies()
    for line in config.project_config.pretty_dependencies(dependencies):
        click.echo(line)


@click.command(name='list', help="List projects and their locations")
@pass_config
def project_list(config):
    pass


@click.command(name='cd', help="Change to the project's directory")
@pass_config
def project_cd(config):
    pass

project.add_command(project_init)
project.add_command(project_info)
project.add_command(project_dependencies)
# project.add_command(project_list)
# project.add_command(project_cd)

# Commands for group: service


@click.command(name='list', help='List services available for configuration and use')
@pass_config
def service_list(config):
    headers = ['service', 'description', 'is_configured']
    data = []
    for serv, schema in list(config.project_config.services.items()):
        is_configured = ''
        if serv in config.keychain.list_services():
            is_configured = '* '
        data.append((serv, schema['description'], is_configured))
    table = Table(data, headers)
    click.echo(table)


class ConnectServiceCommand(click.MultiCommand):

    def list_commands(self, ctx):
        """ list the services that can be configured """
        config = ctx.ensure_object(CliConfig)
        return sorted(config.project_config.services.keys())

    def _build_param(self, attribute, details):
        req = details['required']
        return click.Option(('--{0}'.format(attribute),), prompt=req, required=req)

    def get_command(self, ctx, name):
        config = ctx.ensure_object(CliConfig)

        attributes = iter(list(getattr(
            config.project_config,
            'services__{0}__attributes'.format(name)
        ).items()))
        params = [self._build_param(attr, cnfg) for attr, cnfg in attributes]
        params.append(click.Option(('--project',), is_flag=True))

        @click.pass_context
        def callback(ctx, project=False, *args, **kwargs):
            check_keychain(config)
            serv_conf = dict((k, v) for k, v in list(kwargs.items())
                             if v != None)  # remove None values
            config.keychain.set_service(
                name, ServiceConfig(serv_conf), project)
            if project:
                click.echo(
                    '{0} is now configured for this project'.format(name))
            else:
                click.echo('{0} is now configured for global use'.format(name))

        ret = click.Command(name, params=params, callback=callback)
        return ret


@click.command(cls=ConnectServiceCommand, name='connect', help='Connect a CumulusCI task service')
@click.pass_context
def service_connect(ctx, *args, **kvargs):
    pass


@click.command(name='show', help='Show the details of a connected service')
@click.argument('service_name')
@pass_config
def service_show(config, service_name):
    check_keychain(config)
    try:
        service_config = config.keychain.get_service(service_name)
        click.echo(render_recursive(service_config.config))
    except ServiceNotConfigured:
        click.echo('{0} is not configured for this project.  Use service connect {0} to configure.'.format(
            service_name))

service.add_command(service_connect)
service.add_command(service_list)
service.add_command(service_show)

# Commands for group: org


@click.command(name='browser', help="Opens a browser window and logs into the org using the stored OAuth credentials")
@click.argument('org_name')
@pass_config
def org_browser(config, org_name):
    check_connected_app(config)

    org_config = config.project_config.get_org(org_name)
    org_config = check_org_expired(config, org_name, org_config)
    org_config.refresh_oauth_token(config.keychain.get_connected_app())

    webbrowser.open(org_config.start_url)

    # Save the org config in case it was modified
    config.keychain.set_org(org_config)


@click.command(name='connect', help="Connects a new org's credentials using OAuth Web Flow")
@click.argument('org_name')
@click.option('--sandbox', is_flag=True, help="If set, connects to a Salesforce sandbox org")
@click.option('--login-url', help='If set, login to this hostname.', default='https://login.salesforce.com')
@click.option('--default', is_flag=True, help='If set, sets the connected org as the new default org')
@click.option('--global-org', help='Set True if org should be used by any project', is_flag=True)
@pass_config
def org_connect(config, org_name, sandbox, login_url, default, global_org):
    check_connected_app(config)
    check_org_overwrite(config, org_name)

    connected_app = config.keychain.get_connected_app()
    if sandbox:
        login_url = 'https://test.salesforce.com'

    oauth_capture = CaptureSalesforceOAuth(
        client_id=connected_app.client_id,
        client_secret=connected_app.client_secret,
        callback_url=connected_app.callback_url,
        auth_site=login_url,
        scope='web full refresh_token'
    )
    oauth_dict = oauth_capture()
    org_config = OrgConfig(oauth_dict, org_name)
    org_config.load_userinfo()

    config.keychain.set_org(org_config, global_org)

    if default:
        org = config.keychain.set_default_org(org_name)
        click.echo('{} is now the default org'.format(org_name))


@click.command(name='default', help="Sets an org as the default org for tasks and flows")
@click.argument('org_name')
@click.option('--unset', is_flag=True, help="Unset the org as the default org leaving no default org selected")
@pass_config
def org_default(config, org_name, unset):
    check_connected_app(config)

    if unset:
        org = config.keychain.unset_default_org()
        click.echo(
            '{} is no longer the default org.  No default org set.'.format(org_name))
    else:
        org = config.keychain.set_default_org(org_name)
        click.echo('{} is now the default org'.format(org_name))


@click.command(name='info', help="Display information for a connected org")
@click.argument('org_name')
@click.option('print_json', '--json', is_flag=True, help="Print as JSON")
@pass_config
def org_info(config, org_name, print_json):
    check_connected_app(config)

    org_config = config.keychain.get_org(org_name)
    org_config = check_org_expired(config, org_name, org_config)

    try:
        org_config.refresh_oauth_token(config.keychain.get_connected_app())
    except ScratchOrgException as e:
        raise click.ClickException('ScratchOrgException: {}'.format(e.message))

    if print_json:
        click.echo(json.dumps(org_config.config, sort_keys=True, indent=4))
    else:
        click.echo(render_recursive(org_config.config))

    if org_config.scratch and org_config.expires:
        click.echo('Org expires on {:%c}'.format(org_config.expires))
        
    # Save the org config in case it was modified
    config.keychain.set_org(org_config)


@click.command(name='list', help="Lists the connected orgs for the current project")
@pass_config
def org_list(config):
    check_connected_app(config)
    data = []
    headers = ['org', 'default', 'scratch', 'days', 'expired', 'config_name', 'username']
    for org in config.project_config.list_orgs():
        org_config = config.project_config.get_org(org)
        row = [org]
        row.append('*' if org_config.default else '')
        row.append('*' if org_config.scratch else '')
        if org_config.days_alive:
            row.append('{} of {}'.format(org_config.days_alive, org_config.days) if org_config.scratch else '')
        else:
            row.append(org_config.days if org_config.scratch else '')
        row.append('*' if org_config.scratch and org_config.expired else '')
        row.append(org_config.config_name if org_config.config_name else '')
        username = org_config.config.get('username', org_config.userinfo__preferred_username)
        row.append(username if username else '')
        data.append(row)
    table = Table(data, headers)
    click.echo(table)

@click.command(name='remove', help="Removes an org from the keychain")
@click.argument('org_name')
@click.option('--global-org', is_flag=True, help="Set this option to force remove a global org.  Default behavior is to error if you attempt to delete a global org.")
@pass_config
def org_remove(config, org_name, global_org):
    check_connected_app(config)

    try:
        org_config = config.keychain.get_org(org_name)
        if isinstance(org_config, ScratchOrgConfig):
            if org_config.date_created:
                # attempt to delete the org
                try:
                    click.echo("A scratch org was already created, attempting to delete...")
                    org_config.delete_org()
                except Exception as e:
                    click.echo("Deleting scratch org failed with error:")
                    click.echo(e)
    except OrgNotFound:
        raise click.ClickException('Org {} does not exist in the keychain'.format(org_name))
    config.keychain.remove_org(org_name, global_org)

@click.command(name='scratch', help="Connects a Salesforce DX Scratch Org to the keychain")
@click.argument('config_name')
@click.argument('org_name')
@click.option('--default', is_flag=True, help='If set, sets the connected org as the new default org')
@click.option('--delete', is_flag=True, help="If set, triggers a deletion of the current scratch org.  This can be used to reset the org as the org configuration remains to regenerate the org on the next task run.")
@click.option('--devhub', help="If provided, overrides the devhub used to create the scratch org")
@click.option('--days', help="If provided, overrides the scratch config default days value for how many days the scratch org should persist")
@pass_config
def org_scratch(config, config_name, org_name, default, delete, devhub, days):
    check_connected_app(config)
    check_org_overwrite(config, org_name)

    scratch_configs = getattr(config.project_config, 'orgs__scratch')
    if not scratch_configs:
        raise click.UsageError('No scratch org configs found in cumulusci.yml')
    scratch_config = scratch_configs.get(config_name)
    if not scratch_config:
        raise click.UsageError(
            'No scratch org config named {} found in the cumulusci.yml file'.format(
                config_name)
        )

    if devhub:
        scratch_config['devhub'] = devhub

    config.keychain.create_scratch_org(org_name, config_name, days)

    if default:
        org = config.keychain.set_default_org(org_name)
        click.echo('{} is now the default org'.format(org_name))


@click.command(name='scratch_delete', help="Deletes a Salesforce DX Scratch Org leaving the config in the keychain for regeneration")
@click.argument('org_name')
@pass_config
def org_scratch_delete(config, org_name):
    check_connected_app(config)
    org_config = config.keychain.get_org(org_name)
    if not org_config.scratch:
        raise click.UsageError('Org {} is not a scratch org'.format(org_name))

    try:
        org_config.delete_org()
    except ScratchOrgException as e:
        exception = click.UsageError(e.message)

    config.keychain.set_org(org_config)


@click.command(name='connected_app', help="Displays the ConnectedApp info used for OAuth connections")
@pass_config
def org_connected_app(config):
    check_connected_app(config)
    click.echo(render_recursive(config.keychain.get_connected_app().config))


@click.command(name='config_connected_app', help="Configures the connected app used for connecting to Salesforce orgs")
@click.option('--client_id', help="The Client ID from the connected app", prompt=True)
@click.option('--client_secret', help="The Client Secret from the connected app", prompt=True, hide_input=(False if os.name == 'nt' else True))
@click.option('--callback_url', help="The callback_url configured on the Connected App", default='http://localhost:8080/callback')
@click.option('--project', help='Set if storing encrypted keychain file in project directory', is_flag=True)
@pass_config
def org_config_connected_app(config, client_id, client_secret, callback_url, project):
    check_keychain(config)
    app_config = ConnectedAppOAuthConfig()
    app_config.config = {
        'client_id': client_id,
        'client_secret': client_secret,
        'callback_url': callback_url,
    }
    config.keychain.set_connected_app(app_config, project)

org.add_command(org_browser)
org.add_command(org_config_connected_app)
org.add_command(org_connect)
org.add_command(org_connected_app)
org.add_command(org_default)
org.add_command(org_info)
org.add_command(org_list)
org.add_command(org_remove)
org.add_command(org_scratch)
org.add_command(org_scratch_delete)

# Commands for group: task


@click.command(name='list', help="List available tasks for the current context")
@pass_config
def task_list(config):
    check_project_config(config)
    data = []
    headers = ['task', 'description']
    for task in config.project_config.list_tasks():
        data.append((task['name'], task['description']))
    table = Table(data, headers)
    click.echo(table)


@click.command(name='doc', help="Exports RST format documentation for all tasks")
@pass_config
def task_doc(config):
    config_src = config.global_config

    for name, options in list(config_src.tasks.items()):
        task_config = TaskConfig(options)
        doc = doc_task(name, task_config)
        click.echo(doc)
        click.echo('')


@click.command(name='info', help="Displays information for a task")
@click.argument('task_name')
@pass_config
def task_info(config, task_name):
    check_project_config(config)
    task_config = getattr(config.project_config, 'tasks__{}'.format(task_name))
    if not task_config:
        raise TaskNotFoundError('Task not found: {}'.format(task_name))

    task_config = TaskConfig(task_config)
    click.echo(rst2ansi(doc_task(task_name, task_config)))


@click.command(name='run', help="Runs a task")
@click.argument('task_name')
@click.option('--org', help="Specify the target org.  By default, runs against the current default org")
@click.option('-o', nargs=2, multiple=True, help="Pass task specific options for the task as '-o option value'.  You can specify more than one option by using -o more than once.")
@click.option('--debug', is_flag=True, help="Drops into pdb, the Python debugger, on an exception")
@click.option('--debug-before', is_flag=True, help="Drops into the Python debugger right before task start.")
@click.option('--debug-after', is_flag=True, help="Drops into the Python debugger at task completion.")
@click.option('--no-prompt', is_flag=True, help="Disables all prompts.  Set for non-interactive mode use such as calling from scripts or CI systems")
@pass_config
def task_run(config, task_name, org, o, debug, debug_before, debug_after, no_prompt):
    # Check environment
    check_keychain(config)

    # Get necessary configs
    if org:
        org_config = config.project_config.get_org(org)
    else:
        org, org_config = config.project_config.keychain.get_default_org()
    if org_config:
        org_config = check_org_expired(config, org, org_config)
    task_config = getattr(config.project_config, 'tasks__{}'.format(task_name))
    if not task_config:
        raise TaskNotFoundError('Task not found: {}'.format(task_name))

    # Get the class to look up options
    class_path = task_config.get('class_path')
    task_class = import_class(class_path)

    # Parse command line options and add to task config
    if o:
        if 'options' not in task_config:
            task_config['options'] = {}
        for option in o:
            name = option[0]
            value = option[1]

            # Validate the option
            if name not in task_class.task_options:
                raise click.UsageError(
                    'Option "{}" is not available for task {}'.format(
                        name,
                        task_name,
                    ),
                )

            # Override the option in the task config
            task_config['options'][name] = value

    task_config = TaskConfig(task_config)
    exception = None

    # Create and run the task
    try:
        task = task_class(config.project_config,
                          task_config, org_config=org_config)
    except TaskRequiresSalesforceOrg as e:
        exception = click.UsageError(
            'This task requires a salesforce org.  Use org default <name> to set a default org or pass the org name with the --org option')
    except TaskOptionsError as e:
        exception = click.UsageError(e.message)
        handle_exception_debug(config, debug, e, throw_exception=exception)
    except Exception as e:
        handle_exception_debug(config, debug, e, no_prompt=no_prompt)

    if debug_before:
        import pdb
        pdb.set_trace()

    if not exception:
        try:
            task()
        except TaskOptionsError as e:
            exception = click.UsageError(e.message)
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except ApexTestException as e:
            exception = click.ClickException('Failed: ApexTestFailure')
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except BrowserTestFailure as e:
            exception = click.ClickException('Failed: BrowserTestFailure')
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except MetadataComponentFailure as e:
            exception = click.ClickException(
                'Failed: MetadataComponentFailure'
            )
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except MetadataApiError as e:
            exception = click.ClickException('Failed: MetadataApiError')
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except ScratchOrgException as e:
            exception = click.ClickException(
                'ScratchOrgException: {}'.format(e.message))
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except Exception as e:
            handle_exception_debug(config, debug, e, no_prompt=no_prompt)

    if debug_after:
        import pdb
        pdb.set_trace()

    if exception:
        handle_sentry_event(config, no_prompt)
        raise exception


# Add the task commands to the task group
task.add_command(task_doc)
task.add_command(task_info)
task.add_command(task_list)
task.add_command(task_run)

# Commands for group: flow


@click.command(name='list', help="List available flows for the current context")
@pass_config
def flow_list(config):
    check_project_config(config)
    data = []
    headers = ['flow', 'description']
    for flow in config.project_config.flows:
        description = getattr(config.project_config,
                              'flows__{}__description'.format(flow))
        data.append((flow, description))
    table = Table(data, headers)
    click.echo(table)


@click.command(name='info', help="Displays information for a flow")
@click.argument('flow_name')
@pass_config
def flow_info(config, flow_name):
    check_project_config(config)
    flow = getattr(config.project_config, 'flows__{}'.format(flow_name))
    if not flow:
        raise FlowNotFoundError('Flow not found: {}'.format(flow_name))
    click.echo(render_recursive(flow))


@click.command(name='run', help="Runs a flow")
@click.argument('flow_name')
@click.option('--org', help="Specify the target org.  By default, runs against the current default org")
@click.option('--delete-org', is_flag=True, help="If set, deletes the scratch org after the flow completes")
@click.option('--debug', is_flag=True, help="Drops into pdb, the Python debugger, on an exception")
@click.option('-o', nargs=2, multiple=True, help="Pass task specific options for the task as '-o taskname__option value'.  You can specify more than one option by using -o more than once.")
@click.option('--skip', multiple=True, help="Specify task names that should be skipped in the flow.  Specify multiple by repeating the --skip option")
@click.option('--no-prompt', is_flag=True, help="Disables all prompts.  Set for non-interactive mode use such as calling from scripts or CI systems")
@pass_config
def flow_run(config, flow_name, org, delete_org, debug, o, skip, no_prompt):
    # Check environment
    check_keychain(config)

    # Get necessary configs
    if org:
        org_config = config.project_config.get_org(org)
    else:
        org, org_config = config.project_config.keychain.get_default_org()

    org_config = check_org_expired(config, org, org_config)
    
    if delete_org and not org_config.scratch:
        raise click.UsageError(
            '--delete-org can only be used with a scratch org')

    flow = getattr(config.project_config, 'flows__{}'.format(flow_name))
    if not flow:
        raise FlowNotFoundError('Flow not found: {}'.format(flow_name))
    flow_config = FlowConfig(flow)
    if not flow_config.config:
        raise click.UsageError(
            'No configuration found for flow {}'.format(flow_name))

    # Get the class to look up options
    class_path = flow_config.config.get(
        'class_path', 'cumulusci.core.flows.BaseFlow')
    flow_class = import_class(class_path)

    exception = None

    # Parse command line options and add to task config
    options = {}
    if o:
        for option in o:
            options[option[0]] = option[1]

    # Create the flow and handle initialization exceptions
    try:
        flow = flow_class(config.project_config, flow_config,
                          org_config, options, skip)
    except TaskRequiresSalesforceOrg as e:
        exception = click.UsageError(
            'This flow requires a salesforce org.  Use org default <name> to set a default org or pass the org name with the --org option')
    except TaskOptionsError as e:
        exception = click.UsageError(e.message)
        handle_exception_debug(config, debug, e, throw_exception=exception)
    except Exception as e:
        handle_exception_debug(config, debug, e, no_prompt=no_prompt)

    if not exception:
        # Run the flow and handle exceptions
        try:
            flow()
        except TaskOptionsError as e:
            exception = click.UsageError(e.message)
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except ApexTestException as e:
            exception = click.ClickException('Failed: ApexTestException')
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except BrowserTestFailure as e:
            exception = click.ClickException('Failed: BrowserTestFailure')
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except MetadataComponentFailure as e:
            exception = click.ClickException(
                'Failed: MetadataComponentFailure'
            )
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except MetadataApiError as e:
            exception = click.ClickException('Failed: MetadataApiError')
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except ScratchOrgException as e:
            exception = click.ClickException(
                'ScratchOrgException: {}'.format(e.message)
            )
            handle_exception_debug(config, debug, e, throw_exception=exception)
        except Exception as e:
            handle_exception_debug(config, debug, e, no_prompt=no_prompt)

    # Delete the scratch org if --delete-org was set
    if delete_org:
        try:
            org_config.delete_org()
        except Exception as e:
            click.echo(
                'Scratch org deletion failed.  Ignoring the error below to complete the flow:')
            click.echo(e.message)

    if exception:
        handle_sentry_event(config, no_prompt)
        raise exception

flow.add_command(flow_list)
flow.add_command(flow_info)
flow.add_command(flow_run)



from datetime import datetime

from cumulusci.core.exceptions import GithubException
from cumulusci.tasks.github.base import BaseGithubTask

class CloneTag(BaseGithubTask):

    task_options = {
        'src_tag': {
            'description': 'The source tag to clone.  Ex: beta/1.0-Beta_2',
            'required': True,
        },
        'tag': {
            'description': 'The new tag to create by cloning the src tag.  Ex: release/1.0',
            'required': True,
        },
    }
    
    def _run_task(self):
        repo = self.get_repo()
        ref = repo.ref('tags/{}'.format(self.options['src_tag']))
        src_tag = repo.tag(ref.object.sha)
        if not src_tag:
            message = 'Tag {} not found'.format(self.options['src_tag'])
            logger.error(message)
            raise GithubException(message)

        tag = repo.create_tag(
            tag = self.options['tag'],
            message = 'Cloned from {}'.format(self.options['src_tag']),
            sha = src_tag.sha,
            obj_type = 'commit',
            tagger = {
                'name': self.github_config.username,
                'email': self.github_config.email,
                'date': '{}Z'.format(datetime.utcnow().isoformat()),
            },
        )
        self.logger.info('Tag {} created by cloning {}'.format(self.options['tag'], self.options['src_tag']))

        return tag




lex_locators = {
    'app_launcher': {
        'app_link': "//div[@class='slds-section slds-is-open salesforceIdentityAppLauncherDesktopInternal']//section[@id='cards']//div[@class='appTileTitle' and text()='{}']",
        'button': "css: nav.appLauncher button div.slds-icon-waffle",
        'current_app': "//div[contains(@class,'navLeft')]//span[contains(@class,'appName')]/span[text()='{}']",
        'tab_link': "css: a.app-launcher-link[title='{}']",
    },
    'desktop_rendered': 'css: div.desktop.container.oneOne.oneAppLayoutHost[data-aura-rendered-by]',
    'loading_box': 'css: div.auraLoadingBox.oneLoadingBox',
    'modal': {
        'is_open': "css: div.DESKTOP.uiModal.forceModal.open.active",
        'button': "css: div.uiModal div.modal-footer button[title='{}']",
        'has_error': "css: div.pageLevelErrors",
        'error_messages': "css: div.pageLevelErrors ul.errorsList li",
    },
    'object': {
        'button': "css: div.windowViewMode-normal ul.forceActionsContainer.oneActionsRibbon a[title='{}']",
        'field': "//div[contains(@class, 'uiInput')][.//label[contains(@class, 'uiLabel')][.//span[text()='{}']]]//input",
        'field_lookup_value': "//a[@role='option'][.//div[@title='{}']]",
        'record_type_option': "//div[contains(@class, 'changeRecordTypeOptionRightColumn')]//span[text()='{}']",
    },
    'record': {
        'header': {
            'field': "//li[contains(@class, 'slds-page-header__detail')][.//p[contains(@class, 'slds-text-heading--label')][@title='{}']",
            'field_value': "//li[contains(@class, 'slds-page-header__detail')][.//p[contains(@class, 'slds-text-heading--label')][@title='{}']//span[contains(@class, 'uiOutput')][text()]",
            'field_value_link': "//li[contains(@class, 'slds-page-header__detail')][.//p[contains(@class, 'slds-text-heading--label')][@title='{}']]//a",
            'field_value_checked': "//li[contains(@class, 'slds-page-header__detail')][.//p[contains(@class, 'slds-text-heading--label')][@title='{}']]//span[contains(@class, 'uiOutputCheckbox')]//img[@alt='True']",
            'field_value_unchecked': "//li[contains(@class, 'slds-page-header__detail')][.//p[contains(@class, 'slds-text-heading--label')][@title='{}']]//span[contains(@class, 'uiOutputCheckbox')]//img[@alt='False']",
        },
        'related': {
            'button': "//article[contains(@class, 'forceRelatedListCardDesktop')][.//img][.//span[@title='{}']]//a[@title='{}']",
            'count': "//article[contains(@class, 'forceRelatedListCardDesktop')][.//img]//span[@title='{}']/following-sibling::span",
        },
    },
    'spinner': 'css: div.slds-spinner',
    'tabs': {
        'tab': '',
    },
}



""" Utilities for CumulusCI Core

import_class: task class defn import helper
process_bool_arg: determine true/false for a commandline arg
decode_to_unicode: get unicode string from sf api """
from __future__ import unicode_literals

from past.builtins import basestring
def import_class(path):
    """ Import a class from a string module class path """
    components = path.split('.')
    module = components[:-1]
    module = '.'.join(module)
    # __import__ needs a native str() on py2
    mod = __import__(module, fromlist=[str(components[-1])])
    return getattr(mod, str(components[-1]))


def process_bool_arg(arg):
    """ Determine True/False from argument """
    if isinstance(arg, bool):
        return arg
    elif isinstance(arg, basestring):
        if arg.lower() in ['true', '1']:
            return True
        elif arg.lower() in ['false', '0']:
            return False


def decode_to_unicode(content):
    """ decode ISO-8859-1 to unicode, when using sf api """
    if content:
        try:
            # Try to decode ISO-8859-1 to unicode
            return content.decode('ISO-8859-1')
        except UnicodeEncodeError:
            # Assume content is unicode already
            return content



from github3 import login

from cumulusci.core.tasks import BaseTask

class BaseGithubTask(BaseTask):

    def _init_task(self):
        self.github_config = self.project_config.keychain.get_service('github')
        self.github = login(
            username=self.github_config.username,
            password=self.github_config.password,
        )

    def get_repo(self):
        return self.github.repository(
            self.project_config.repo_owner,
            self.project_config.repo_name,
        )



import os
import shutil
import time
import glob
from xml.dom.minidom import parse

from cumulusci.core.tasks import BaseTask
from cumulusci.utils import download_extract_zip


class DownloadZip(BaseTask):
    name = 'Download'
    task_options = {
        'url': {
            'description': 'The url of the zip file to download',
            'required': True,
        },
        'dir': {
            'description': 'The directory where the zip should be extracted',
            'required': True,
        },
        'subfolder': {
            'description': (
                'The subfolder of the target zip to extract. Defaults to' +
                ' extracting the root of the zip file to the destination.'
            ),
        },
    }

    def _run_task(self):
        if not self.options['dir']:
            self.options['dir'] = '.'
        elif not os.path.exists(self.options['dir']):
            os.makedirs(self.options['dir'])

        download_extract_zip(
            self.options['url'],
            self.options['dir'],
            self.options.get('subfolder'),
        )


class ListMetadataTypes(BaseTask):
    name = 'ListMetadataTypes'
    task_options = {
        'package_xml': {'description': (
            'The project package.xml file.' +
            ' Defaults to <project_root>/src/package.xml'
        )}
    }

    def _init_options(self, kwargs):
        super(ListMetadataTypes, self)._init_options(kwargs)
        if 'package_xml' not in self.options:
            self.options['package_xml'] = os.path.join(
                self.project_config.repo_root,
                'src',
                'package.xml',
            )

    def _run_task(self):
        dom = parse(self.options['package_xml'])
        package = dom.getElementsByTagName('Package')[0]
        types = package.getElementsByTagName('types')
        type_list = []
        for t in types:
            name = t.getElementsByTagName('name')[0]
            metadata_type = name.firstChild.nodeValue
            type_list.append(metadata_type)
        self.logger.info(
            'Metadata types found in %s:\r\n%s',
            self.options['package_xml'],
            '\r\n'.join(type_list),
        )


class Sleep(BaseTask):
    name = 'Sleep'
    task_options = {
        'seconds': {
            'description': 'The number of seconds to sleep',
            'required': True,
        },
    }

    def _run_task(self):
        self.logger.info(
            'Sleeping for {} seconds'.format(self.options['seconds'])
        )
        time.sleep(float(self.options['seconds']))
        self.logger.info('Done')

class Delete(BaseTask):
    name = 'Delete'
    task_options = {
        'path': {
            'description': 'The path to delete.  If path is a directory, recursively deletes the directory: BE CAREFUL!!!  If path is a list, all paths will be deleted',
            'required': True,
        },
        'chdir': {
            'description': 'Change directories before deleting path(s).  This is useful if you have a common list of relative paths to delete that you want to call against different directories.',
        }
    }

    def _run_task(self):
        chdir = self.options.get('chdir')
        cwd = os.getcwd()
        if chdir:
            self.logger.info(
                'Changing directory to {}'.format(chdir)
            )
            os.chdir(chdir)

        path = self.options['path']
        if isinstance(path, list):
            for path_item in path:
                for match in glob.glob(path_item):
                    self._delete(match)
   
        if chdir: 
            os.chdir(cwd)

    def _delete(self, path):
        if not os.path.exists(path):
            self.logger.info(
                '{} does not exist, skipping delete'.format(path)
            )
        if os.path.isdir(path):
            self.logger.info(
                'Recursively deleting directory {}'.format(path)
            )
            shutil.rmtree(path)
        else:
            self.logger.info(
                'Deleting file {}'.format(path)
            )
            os.remove(path)


class FindReplace(BaseTask):
    task_options = {
        'find': {
            'description': "The string to search for",
            'required': True,
        },
        'replace': {
            'description': "The string to replace matches with. Defaults to an empty string",
            'required': True,
        },
        'path': {
            'description': "The path to recursively search",
            'required': True,
        },
        'file_pattern': {
            'description': "A UNIX like filename pattern used for matching filenames.  See python fnmatch docs for syntax.  Defaults to *",
            'required': True,
        },
        'max': {
            'description': "The max number of matches to replace.  Defaults to replacing all matches.",
        },
    }
    
    def _init_options(self):
        if 'replace' not in self.options:
            self.options['replace'] = ''
        if 'file_pattern' not in self.options:
            self.options['file_pattern'] = '*'

    def _run_task(self):
        kwargs = {}
        if 'max' in self.options:
            kwargs['max'] = self.options['max']
        findReplace(
            find = self.options['find'],
            replace = self.options['replace'],
            directory = self.options['path'],
            filePattern = self.options['file_pattern'],
            logger = self.logger,
            **kwargs
        )

find_replace_regex_options = FindReplace.task_options.copy()
del find_replace_regex_options['max']

class FindReplaceRegex(FindReplace):
    task_options = find_replace_regex_options
            
    def _run_task(self):
        findReplaceRegex(
            find = self.options['find'],
            replace = self.options['replace'],
            directory = self.options['path'],
            filePattern = self.options['file_pattern'],
            logger = self.logger,
        )

class CopyFile(BaseTask):
    task_options = {
        'src': {
            'description': 'The path to the source file to copy',
            'required': True,
        },
        'dest': {
            'description': 'The destination path where the src file should be copied',
            'required': True,
        },
    }

    def _run_task(self):
        self.logger.info('Copying file {src} to {dest}'.format(**self.options))
        shutil.copyfile(
            src = self.options['src'],
            dst = self.options['dest'],
        )



import ast
import httplib
import os
import shutil
import tempfile
import unittest

import requests
import responses

from datetime import datetime
from datetime import timedelta
from testfixtures import LogCapture

from cumulusci.tests.util import create_project_config
from cumulusci.tests.util import DummyOrgConfig
from cumulusci.core.config import ServiceConfig
from cumulusci.core.config import TaskConfig

from cumulusci.tasks.github import MergeBranch
from cumulusci.core.exceptions import GithubApiNotFoundError
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin


class TestMergeBranch(unittest.TestCase, GithubApiTestMixin):

    def setUp(self):

        # Set up the mock values
        self.repo_name = 'TestRepo'
        self.repo_owner = 'TestOwner'
        self.repo_api_url = 'https://api.github.com/repos/{}/{}'.format(
            self.repo_owner,
            self.repo_name,
        )
        self.branch = 'master'

        # Create the project config
        self.project_config = create_project_config(
            self.repo_name,
            self.repo_owner,
        )
        self.project_config.keychain.set_service(
            'github',
            ServiceConfig({ 
                'username': 'TestUser',
                'password': 'TestPass',
                'email': 'testuser@testdomain.com',
            })
        )
        
        #self.current_tag_sha = self._random_sha()
        #self.current_tag_commit_sha = self._random_sha()
        #self.current_tag_commit_date = datetime.utcnow()
        #self.last_tag_sha = self._random_sha()
        #self.last_tag_commit_sha = self._random_sha()
        #self.last_tag_commit_date = datetime.utcnow() - timedelta(days=1)

    def _create_task(self, task_config=None):
        if not task_config:
            
            task_config = {}
        task = MergeBranch(
            project_config = self.project_config,
            task_config = TaskConfig(task_config),
            org_config = DummyOrgConfig(),
        )
        return task

    def _mock_repo(self):
        api_url = self.repo_api_url
        
        expected_response = self._get_expected_repo(
            owner = self.repo_owner,
            name = self.repo_name,
        )
        responses.add(
            method=responses.GET,
            url=api_url,
            json=expected_response,
        )
        return expected_response

    def _mock_branch(self, branch, expected_response=None):
        api_url = '{}/branches/{}'.format(self.repo_api_url, branch)
        if not expected_response:
            expected_response = self._get_expected_branch(branch)
        responses.add(
            method = responses.GET,
            url = api_url,
            json = expected_response,
        )

    def _mock_branches(self, branches=None):
        api_url = '{}/branches'.format(self.repo_api_url)
        if branches:
            expected_response = branches
        else:
            expected_response = []
        
        master = self._get_expected_branch(
            'master', 
            self.project_config.repo_commit
        )
        expected_response = [master] + expected_response

        responses.add(
            method = responses.GET,
            url = api_url,
            json = expected_response,
        )
        return expected_response

    def _mock_branch_does_not_exist(self, branch):
        api_url = '{}/branches/{}'.format(self.repo_api_url, branch)
        expected_response = self._get_expected_not_found()
        responses.add(
            method = responses.GET,
            url = api_url,
            status = httplib.NOT_FOUND, # 404
            json = expected_response,
        )

    def _mock_merge(self, conflict=None):
        api_url = '{}/merges'.format(self.repo_api_url)
        expected_response = self._get_expected_merge(conflict)
        status = httplib.CREATED # 201
        if conflict:
            status = httplib.CONFLICT # 409
        
        responses.add(
            method = responses.POST,
            url = api_url,
            json = expected_response,
            status = status,
        )
        return expected_response

    def _mock_pull_create(self, pull_id, issue_id):
        api_url = '{}/pulls'.format(self.repo_api_url)
        expected_response = self._get_expected_pull_request(pull_id, issue_id)
        
        responses.add(
            method = responses.POST,
            url = api_url,
            json = expected_response,
            status = httplib.CREATED, # 201
        )

    def _mock_pulls(self, pulls=None):
        api_url = '{}/pulls'.format(self.repo_api_url)
        expected_response = self._get_expected_pulls(pulls=pulls)
        
        responses.add(
            method = responses.GET,
            url = api_url,
            json = expected_response,
        )

    def _mock_compare(self, base, head, files=None):
        api_url = '{}/compare/{}...{}'.format(self.repo_api_url, base, head)
        expected_response = self._get_expected_compare(base, head, files)
        
        responses.add(
            method = responses.GET,
            url = api_url,
            json = expected_response,
        )

    def _get_log_lines(self, log):
        log_lines = []
        for event in log.records:
            if event.name != 'cumulusci.core.tasks':
                continue
            log_lines.append((event.levelname, event.getMessage()))
        return log_lines
        

    @responses.activate
    def test_branch_does_not_exist(self):
        self._mock_repo()
        self._mock_branch_does_not_exist(self.branch)

        task = self._create_task()
        with self.assertRaises(GithubApiNotFoundError):
            task()
        self.assertEquals(2, len(responses.calls))

    @responses.activate
    def test_no_feature_branch(self):
        self._mock_repo()
        self._mock_branch(self.branch)
        other_branch = self._get_expected_branch('not-a-feature-branch')
        self._mock_pulls()
        branches = [other_branch]
        branches = self._mock_branches(branches)
        with LogCapture() as l:
            task = self._create_task()
            task()
            log_lines = self._get_log_lines(l)
                    
            expected = [
                ('INFO', 'Beginning task: MergeBranch'),
                ('INFO', ''),
                ('DEBUG', 'Skipping branch master: is source branch'),
                ('DEBUG', 'Skipping branch not-a-feature-branch: does not match prefix feature/'),
            ]
            self.assertEquals(expected, log_lines)
        self.assertEquals(4, len(responses.calls))

    @responses.activate
    def test_feature_branch_no_diff(self):
        self._mock_repo()
        self._mock_branch(self.branch)
        self._mock_pulls()
        branch_name = 'feature/a-test'
        branches = []
        branches.append(self._get_expected_branch(branch_name))
        branches = self._mock_branches(branches)
        self._mock_compare(
            base = branches[1]['name'],
            head = self.project_config.repo_commit,
        )
        with LogCapture() as l:
            task = self._create_task()
            task()
            log_lines = self._get_log_lines(l)

            expected = [
                ('INFO', 'Beginning task: MergeBranch'),
                ('INFO', ''),
                ('DEBUG', 'Skipping branch master: is source branch'),
                ('INFO', 'Skipping branch feature/a-test: no file diffs found'),
            ]
            self.assertEquals(expected, log_lines)
        self.assertEquals(5, len(responses.calls))

    @responses.activate
    def test_feature_branch_merge(self):
        self._mock_repo()
        self._mock_branch(self.branch)
        self._mock_pulls()
        branch_name = 'feature/a-test'
        branches = []
        branches.append(self._get_expected_branch(branch_name))
        branches = self._mock_branches(branches)
        self._mock_compare(
            base = branches[1]['name'],
            head = self.project_config.repo_commit,
            files = [
                {'filename': 'test.txt'},
            ]
        )
        self._mock_merge()
        with LogCapture() as l:
            task = self._create_task()
            task()
            log_lines = self._get_log_lines(l)

            expected = [
                ('INFO', 'Beginning task: MergeBranch'),
                ('INFO', ''),
                ('DEBUG', 'Skipping branch master: is source branch'),
                ('INFO', 'Merged 1 commits into branch feature/a-test'),
            ]
            self.assertEquals(expected, log_lines)
        self.assertEquals(6, len(responses.calls))
    
    @responses.activate
    def test_feature_branch_merge_conflict(self):
        self._mock_repo()
        self._mock_branch(self.branch)
        self._mock_pulls()
        branch_name = 'feature/a-test'
        branches = []
        branches.append(self._get_expected_branch(branch_name))
        branches = self._mock_branches(branches)
        self._mock_compare(
            base = branches[1]['name'],
            head = self.project_config.repo_commit,
            files = [
                {'filename': 'test.txt'},
            ]
        )
        self._mock_merge(True)
        self._mock_pull_create(1, 2)
        with LogCapture() as l:
            task = self._create_task()
            task()
            log_lines = self._get_log_lines(l)

            expected = [
                ('INFO', 'Beginning task: MergeBranch'),
                ('INFO', ''),
                ('DEBUG', 'Skipping branch master: is source branch'),
                ('INFO', 'Merge conflict on branch feature/a-test: created pull request #2')
            ]
            self.assertEquals(expected, log_lines)
        self.assertEquals(7, len(responses.calls))

    @responses.activate
    def test_feature_branch_existing_pull(self):
        self._mock_repo()
        self._mock_branch(self.branch)

        branch_name = 'feature/a-test'
        branches = []
        branches.append(self._get_expected_branch(branch_name))
        branches = self._mock_branches(branches)

        pull = self._get_expected_pull_request(1, 2)
        pull['base']['ref'] = branch_name
        pull['base']['sha'] = branches[1]['commit']['sha']
        pull['head']['ref'] = self.branch
        self._mock_pulls([pull])

        self._mock_compare(
            base = branches[1]['name'],
            head = self.project_config.repo_commit,
            files = [
                {'filename': 'test.txt'},
            ]
        )
        self._mock_merge(True)

        with LogCapture() as l:
            task = self._create_task()
            task()
            log_lines = self._get_log_lines(l)

            expected = [
                ('INFO', 'Beginning task: MergeBranch'),
                ('INFO', ''),
                ('DEBUG', 'Skipping branch master: is source branch'),
                ('INFO', 'Merge conflict on branch feature/a-test: merge PR already exists'),
            ]
            self.assertEquals(expected, log_lines)
        self.assertEquals(6, len(responses.calls))


    @responses.activate
    def test_master_parent_with_child_pr(self):
        self._mock_repo()
        self._mock_branch(self.branch)
        # branches
        parent_branch_name = 'feature/a-test'
        child_branch_name = 'feature/a-test__a-child'
        branches = []
        branches.append(self._get_expected_branch(parent_branch_name))
        branches.append(self._get_expected_branch(child_branch_name))
        branches = self._mock_branches(branches)
        # pull request
        pull = self._get_expected_pull_request(1, 2)
        pull['base']['ref'] = parent_branch_name
        pull['base']['sha'] = branches[1]['commit']['sha']
        pull['head']['ref'] = child_branch_name
        self._mock_pulls([pull])
        # compare
        self._mock_compare(
            base = parent_branch_name,
            head = self.project_config.repo_commit,
            files = [
                {'filename': 'test.txt'},
            ]
        )
        # merge
        self._mock_merge(True)
        # create PR
        self._mock_pull_create(1, 2)
        with LogCapture() as l:
            task = self._create_task()
            task()
            log_lines = self._get_log_lines(l)
        expected = [
            ('INFO', 'Beginning task: MergeBranch'),
            ('INFO', ''),
            ('DEBUG', 'Skipping branch master: is source branch'),
            ('INFO', 'Merge conflict on parent branch {}: created pull request #2'.format(
                parent_branch_name
            )),
        ]
        self.assertEquals(expected, log_lines)
        self.assertEquals(7, len(responses.calls))

    @responses.activate
    def test_master_parent_does_not_merge_child(self):
        self._mock_repo()
        self._mock_branch(self.branch)

        parent_branch_name = 'feature/a-test'
        child1_branch_name = 'feature/a-test__a-child1'
        child2_branch_name = 'feature/a-test__a-child2'
        branches = []
        branches.append(self._get_expected_branch(parent_branch_name))
        branches.append(self._get_expected_branch(child1_branch_name))
        branches.append(self._get_expected_branch(child2_branch_name))
        branches = self._mock_branches(branches)

        self._mock_pulls()

        self._mock_compare(
            base = branches[1]['name'],
            head = self.project_config.repo_commit,
            files = [
                {'filename': 'test.txt'},
            ]
        )
        merges = []
        merges.append(self._mock_merge())

        with LogCapture() as l:
            task = self._create_task()
            task()

            log_lines = self._get_log_lines(l)

            expected = [
                ('INFO', 'Beginning task: MergeBranch'),
                ('INFO', ''),
                ('DEBUG', 'Skipping branch master: is source branch'),
                ('INFO', 'Merged 1 commits into parent branch {}'.format(
                    parent_branch_name
                )),
                ('INFO', '  Skipping merge into the following child branches:'),
                ('INFO', '    {}'.format(child1_branch_name)),
                ('INFO', '    {}'.format(child2_branch_name)),
            ]
            self.assertEquals(expected, log_lines)
        self.assertEquals(6, len(responses.calls))


    @responses.activate
    def test_parent_merge_to_children(self):
        branch = 'feature/a-test'
        self._mock_repo()
        self._mock_branch(branch)

        parent_branch_name = 'feature/a-test'
        child1_branch_name = 'feature/a-test__a-child1'
        child2_branch_name = 'feature/a-test__a-child2'
        branches = []
        branches.append(self._get_expected_branch(parent_branch_name))
        branches.append(self._get_expected_branch(child1_branch_name))
        branches.append(self._get_expected_branch(child2_branch_name))
        branches = self._mock_branches(branches)

        self._mock_pulls()

        merges = []
        merges.append(self._mock_merge())

        self._mock_compare(
            base = branches[2]['name'],
            head = self.project_config.repo_commit,
            files = [
                {'filename': 'test.txt'},
            ]
        )
        self._mock_compare(
            base = branches[3]['name'],
            head = self.project_config.repo_commit,
            files = []
        )

        with LogCapture() as l:
            task = self._create_task(task_config={
                'options': {
                    'source_branch': 'feature/a-test',
                    'children_only': True,
                }
            })
            task()

            log_lines = self._get_log_lines(l)

            expected = [
                ('INFO', 'Beginning task: MergeBranch'),
                ('INFO', ''),
                (
                    'INFO', 
                    'Performing merge from parent branch {} to children'.format(
                        branches[1]['name'],
                    ),
                ),
                ('INFO', 'Merged 1 commits into child branch {}'.format(
                    branches[2]['name']
                )),
                ('INFO', 'Skipping child branch {}: no file diffs found'.format(
                    branches[3]['name']
                )),
            ]
            self.assertEquals(expected, log_lines)
        self.assertEquals(7, len(responses.calls))

    @responses.activate
    def test_parent_merge_no_children(self):
        branch = 'feature/a-test'
        self._mock_repo()
        self._mock_branch(branch)

        parent_branch_name = 'feature/a-test'
        child1_branch_name = 'feature/b-test'
        branches = []
        branches.append(self._get_expected_branch(parent_branch_name))
        branches.append(self._get_expected_branch(child1_branch_name))
        branches = self._mock_branches(branches)

        self._mock_pulls()

        with LogCapture() as l:
            task = self._create_task(task_config={
                'options': {
                    'source_branch': 'feature/a-test',
                    'children_only': True,
                }
            })
            task()

            log_lines = self._get_log_lines(l)

            expected = [
                ('INFO', 'Beginning task: MergeBranch'),
                ('INFO', ''),
                (
                    'INFO', 
                    'No children found for branch {}'.format(
                        branches[1]['name'],
                    ),
                ),
            ]
            self.assertEquals(expected, log_lines)
        self.assertEquals(4, len(responses.calls))
